  CPU   386  ;Don't allow 486+ Instructions
  USE16      ;16-bit Code

  %include "A86Help.Mac" ;Include A86 to NASM conversion Macros

;##############################################################################
;EQUATES
;##############################################################################

;==============================================================================
;Miscellaneous
;==============================================================================

  ;----------------------------------------------------------------------------
  ;Special Codes
  ;----------------------------------------------------------------------------
  No  EQU  0 ;Miscellaneous Yes/No tests
  Yes EQU -1 ;      "        "  "    "


;==============================================================================
;DOS Drive/File Related
;==============================================================================
  DefSectorSize  EQU   512  ;Default Sector Size.  Technically, Drives can have
                            ;  any value that is an exponent of 2 between 32
                            ;  and 8192, BUT many programs expect only 512.
                            ;  Smaller sectors handled worse than larger ones!
  MinSectorSize  EQU    32  ;Minimum Sector Size allowed by DOS
;  MaxSectorSize  EQU  8192  ;Maximum Sector Size allowed by DOS
  MaxSectorSize  EQU 32768  ;Maximum Sector Size allowed by DOS (only MS-DOS)
  BootRecSig     EQU 0AA55h ;Signature at Offset 510 (last two bytes)
                            ;  of a Partition Record or Boot Record Sector
  MaxFAT16Sects  EQU 4194304 ;Maximum number of sectors in a FAT16 partition
                             ;  (65536 Clusters * 64 Sectors / Cluster)

  ;----------------------------------------------------------------------------
  ;Special CPU OpCodes we need to concern ourselves with
  ;----------------------------------------------------------------------------
  OpCodeJmpNear  EQU 0E9h ;JMP NEAR  (Within Same Segment)
  OpCodeJmpShort EQU 0EBh ;JMP SHORT (+127/-128 Bytes)


;##############################################################################
;STRUCTURES
;##############################################################################

;==============================================================================
;BIOS Related
;==============================================================================

  ;----------------------------------------------------------------------------
  ;Fixed Disk Parameter Structure
  ;Two of these are stored in memory, one for the first hard drive (80h),
  ;  pointed at by INT 41h in the Interrupt Vector Table, and the other
  ;  for the second hard drive (81h), pointed at by INT 46h in the IVT.
  ;Theoretically, modern programs should not ever look at these tables
  ;  directly, but instead should rely on INT 13h functions to tell them
  ;  the disk parameters.  Unfortunately, some programs do not do that.
  ;----------------------------------------------------------------------------
  STRUC FDPTStruc
    FDPTNumTracks     RESW  1  ;Number of Tracks (Cylinders), Max 1024
    FDPTNumHeads      RESB  1  ;Number of Heads, Max 255
    FDPTReducedCyl    RESW  1  ;Starting Reduced Write Current Cylinder (XT Only)
    FDPTPrecompCyl    RESW  1  ;Starting Precompensation cylinder (Obsolete)
    FDPTMaxECCBurst   RESB  1  ;Maximum ECC Burst Length (XT Only)
    FDPTCtlByte       RESB  1  ;Control Byte (following bit definitions are AT+)
      FDPTHeads8Plus   EQU 08h ;More than 8 Heads (bit 0-2, 4 all reserved: 0)
      FDPTDefectMap    EQU 20h ;Mfg Defect Map (Stored on Max Track + 1)
      FDPTNoECCRetry   EQU 40h ;Disable ECC Retries
      FDPTNoAccRetry   EQU 80h ;Disable Access Retries
    FDPTStdTimeout    RESB  1  ;Standard Timeout (XT Only)
    FDPTFmtTimeout    RESB  1  ;Formatting Timeout (XT & WD1002 Only)
    FDPTChkTimeout    RESB  1  ;Check Drive Timeout (XT & WD1002 Only)
    FDPTLandingZone   RESW  1  ;Landing Zone Cylinder (Obsolete)
    FDPTSectsPerTrack RESB  1  ;Number of Sectors Per Track, Max 63
    FDPTPAD           RESB  1  ;PAD/Reserved
   ENDS
   FDPTSize EQU FDPTStruc_size ;16 bytes

  ;----------------------------------------------------------------------------
  ;Enhanced Disk Parameter Structure
  ;This is used for BIOSes that "Translate" Logical CHS (BIOS) to Physical
  ;  CHS (IDE/EIDE).
  ;This is also required for Drives larger than 8 GB, since we need to put
  ;  16383/16/63 in Physical CHS to indicate a "fake" drive size.
  ;----------------------------------------------------------------------------
  STRUC EDPTStruc
    EDPTLNumTracks     RESW  1  ;Logical Number of Tracks, Max 1024
    EDPTLNumHeads      RESB  1  ;Logical Number of Heads, Max 255
    EDPTSignature      RESB  1  ;A0h Signature for EDPT
    EDPTPSectsPerTrack RESB  1  ;Physical Sectors Per Track, Max 255
    EDPTPrecompCyl     RESW  1  ;Starting Precompensation Cylinder (Obsolete)
    EDPTPAD            RESB  1  ;PAD/Reserved
    EDPTCtlByte        RESB  1  ;Control Byte
      ;Same as FDPTCtlByte (FDPTHeads8Plus, etc.)
    EDPTPNumTracks     RESW  1  ;Physical Number of Tracks, Max 65535
    EDPTPNumHeads      RESB  1  ;Physical Number of Heads, Max 16
    EDPTLandingZone    RESW  1  ;Landing Zone Cylinder (Obsolete)
    EDPTLSectsPerTrack RESB  1  ;Logical Sectors Per Track, Max 63
    EDPTCheckSum       RESB  1  ;CheckSum
   ENDS ;16 bytes

;==============================================================================
;Disk-Related
;==============================================================================

  ;----------------------------------------------------------------------------
  ;Structure for a Partition Record
  ;There are four of these in each Partition Record (the first being the
  ;  Master Boot Record on the first Sector of a Hard Disk).  There are also
  ;  four in each Extended Partition Record, if any exist.
  ;----------------------------------------------------------------------------
  STRUC PartRecordStruc
    PartPrimary     RESB   1  ;Primary (Active) Partition Indicator
      PPrimaryNo     EQU  00h ;Is not the Primary (Bootable) Partition
      PPrimaryYes    EQU  80h ;Is the Primary (Bootable) Partition
                              ;There can only be one Primary Partition!
    PartStartHead   RESB   1  ;Partition Starting Head
    PartStartSect   RESB   1  ;Partition Starting Sector
      PSectSector    EQU  3Fh ;Sector is only in bits 0-5
      PSectCyl89     EQU 0C0h ;Bits 8-9 of Starting Cylinder/Track
    PartStartCyl    RESB   1  ;Partition Starting Cylinder/Track
                              ;  (bits 8-9 come from PartStartSect)
    PartOpSys       RESB   1  ;Operating System Indicator
      POSNone        EQU  00h ;None (Empty Parition Entry)
      POSFAT12       EQU  01h ;DOS 12-bit FAT
      POSFAT16       EQU  04h ;FAT16 (max 32M)
      POSDOSExt      EQU  05h ;DOS Extended
      POSFAT16H      EQU  06h ;FAT16 (32M+)
      POSFAT32       EQU  0Bh ;FAT32
      POSFAT32LBA    EQU  0Ch ;FAT32 w/ LBA extensions
      POSLBAVFAT16H  EQU  0Eh ;LBA VFAT (same as 06h but using LBA-mode INT 13)
      POSLBAVFATExt  EQU  0Fh ;LBA VFAT (same as 05h but using LBA-mode INT 13)
                              ;The following OS type list is not guaranteed to
                              ;  be either accurate or complete, other than the
                              ;  low-numbered ones relating to DOS, EE and EF
      POSLinuxNFS    EQU  83h ;Linux Native File System (ext2fs/xiafs)
      POSLinuxExt    EQU  85h ;Linux EXT
      POSLinuxLVM    EQU  8Eh ;Linux LVM
      POSFreeBSD     EQU 0A5h ;FreeBSD, BSD/386, BSD Variant
      POSOpenBSD     EQU 0A6h ;OpenBSD
      POSNetBSD      EQU 0A9h ;NetBSD
      POSDarwin      EQU 0ABh ;Apple Darwin Boot
      POSGPT         EQU 0EEh ;GUID Partition Table
      POSEFIGUID     EQU 0EFh ;EFI GUID (Modern Disks >2TB)
    PartEndHead     RESB   1  ;Partition Ending Head
    PartEndSect     RESB   1  ;Partition Ending Sector (bits 0-5)
    PartEndCyl      RESB   1  ;Partition Ending Cylinder/Track
                              ;  (bits 8-9 come from PartEndSect)
    PartLBASector   RESD   1  ;LBA Starting Sector (number of sectors
                              ;  preceding the partition)
    PartLBALength   RESD   1  ;Length of Partition (Sectors)
   ENDS ;16 bytes
  PRecordSize EQU PartRecordStruc_size ;16 Bytes

  ;----------------------------------------------------------------------------
  ;Structure for a Master Boot Record or Extended Partition Record
  ;----------------------------------------------------------------------------
  STRUC MasterBootStruc
    MBRCode      RESB         446 ;446 bytes of unknown data
                                  ;For MBR, this is the boot code
                                  ;For Extended Partition Record, this is zeroes
    MBRPart1     RESB PRecordSize ;Partition 1 Record
    MBRPart2     RESB PRecordSize ;Partition 2 Record
    MBRPart3     RESB PRecordSize ;Partition 3 Record
    MBRPart4     RESB PRecordSize ;Partition 4 Record
    MBRSignature RESW           1 ;AA55h MBR/PBR Signature
   ENDS ;512 Bytes (1 standard sector)

;==============================================================================
;GPT / GUID Related
;==============================================================================

  ;----------------------------------------------------------------------------
  ;GPT Header Structure
  ;  In Sector 1, immediately following the Protection MBR in Sector 0
  ;  A second copy of this is in the last sector of the disk
  ;----------------------------------------------------------------------------
  STRUC GPTHdrStruc
    GPTHSignature    RESQ 1 ;'TRAP IFE' Signature
    GPTHRevision     RESD 1 ;0001_0000h GPT Revision Number
                            ;  (always v 1.00 through UEFI v2.3.1)
    GPTHHdrSize      RESD 1 ;0000_005Ch Header Size (usually 92 bytes)
    GPTHHdrCRC32     RESD 1 ;CRC32 of Header (0 to Header Size,
                            ;  with this CRC32 field set to 0)
    GPTHPAD1         RESD 1 ;PAD/Reserved (must be 0)
    GPTHThisHdrLBA   RESQ 1 ;LBA64 of This Header (sector 1 if first Header)
    GPTHOtherHdrLBA  RESQ 1 ;LBA64 of Other Header (Last Sector)
    GPTHFirstPartLBA RESQ 1 ;First Usable LBA for Partition Data
                            ;  (usually 34)
    GPTHLastPartLBA  RESQ 1 ;Last Usable LBA for Partition Data
                            ;  (usually DiskSize - 34)
    GPTHHdrGUID      RESQ 2 ;Disk GUID/UUID
    GPTHPartTblLBA   RESQ 1 ;Partition Table Starting LBA (2 if first Header)
    GPTHPartEntries  RESD 1 ;Number of Parition Table Entries (usually 128)
    GPTHPartEntrySz  RESD 1 ;Size of each Partition Table Entry (usually 128)
    GPTHArrayCRC32   RESD 1 ;CRC32 of Partition Array
    ;Rest of Sector is PAD/Reserved (all zeroes, usually 420 bytes)
   ENDS

  ;----------------------------------------------------------------------------
  ;GPT Entry Structure
  ;  Starting in Sector 2, immediately following the GPT Header
  ;  A second copy of this is immediately before the last sector of the disk
  ;----------------------------------------------------------------------------
  STRUC GPTEntryStuc ;GUID Partition Table Entry
    GPTETypeGUID  RESQ 2 ;Partition Type GUID
    GPTEPartGUID  RESQ 2 ;Unique Partition GUID
    GPTEFirstLBA  RESQ 1 ;First LBA (LittleE)
    GPTELastLBA   RESQ 1 ;Last LBA (LittleE)
    GPTEAttribs   RESQ 1 ;Attribute Flags
      ;Bit  0 = System Partition (partitioning utilities must leave as-is)
      ;     2 = Legacy BIOS Bootable (equivalent to MBR 80h Active Flag)
      ;    60 = Read-only, 62 = Hidden, 63 = Do not assign Drive Letter
    GPTEPartName  RESW 36 ;Partition Name (36 UTF-16LE characters)
   ENDS  ;Normally 128 bytes


;==============================================================================
;Partition Processing-Related
;==============================================================================
  ;----------------------------------------------------------------------------
  ;Tracing through all of the partitions to find all appropriate DOS
  ;  Drive Letters is a recursive process, since there is the possibility of
  ;  extended partitions which can contain other extended partitions.
  ;Normally, we would just use the regular stack to be able to trace our way
  ;  up and down the partitions as we are processing them.  Unfortunately,
  ;  we are processing the partitions in the background, and must "stall"
  ;  our processing periodically while we wait for a sector of partition data
  ;  to be transferred from the USB Device into memory where we can use it.
  ;  Therefore, we cannot use the regular stack.
  ;We have to create our own "Partition Stack" in the TSR memory space to
  ;  be able to process things correctly.
  ;----------------------------------------------------------------------------
  STRUC PartStackStruc
    PStkSector RESD 1 ;Sector Number (LBA) of the Partition Record
    PStkOffset RESW 1 ;Offset of Partition Record in SectorBuff
   ENDS
  PStkSize EQU PartStackStruc_size ;6 bytes


;==============================================================================
;Structures:  Drive Parameter Block
;==============================================================================

;==============================================================================
;BIOS Parameter Block
;This is sent back and forth between DOS and all Block Device Drivers
;  (Build BPB call).
;This is also in the Boot Sector of the drive, starting at Offset 11
;  (after three bytes of Code and 8 bytes of OEM Name)
;==============================================================================
  STRUC BIOSParamBlockStruc
    BPBBytesPerSect    RESW   1  ;Number of Bytes per Sector
                                 ;Must be a multiple of 32, "should" be 512!
    BPBSectsPerCluster RESB   1  ;Number of Sectors per Cluster (power of 2)
    BPBReservedSects   RESW   1  ;Number of Reserved Sectors at start of disk
                                 ;  Starting Sector of first FAT (usually 1)
    BPBNumFATs         RESB   1  ;Number of FAT copies (usually 2)
    BPBRootEntries     RESW   1  ;Maximum Number of Root Directory Entries
                                 ;  (0 for FAT32!)
    BPBNumSects16      RESW   1  ;Total Number of Sectors
                                 ;If DOS 4+ and partition > 32MB, set this
                                 ;  to 0 and use DWORD value at Offset 15h
    BPBMediaID         RESB   1  ;Media ID (Descriptor) Byte, for example:
    ; Sides Sectors/Cluster RootEntries TotalSectors Sectors/Track Diam Tracks
    ; FE/FC=160/180+k SS 1 64 320/360+ 8/9+ 5.25 40+,
    ; FF/FD=320/360+k DS 2 112 640/720+ 8/9+ 5.25 40+,
    ; F9=720+k LD 2 112 1440+ 9 3.5 80+ or 1200k HD 1 224 2400 15 5.25 80,
    ; F0=1440+k HD 1 224 2880+ 18 3.5 80+, other F9/F0 variants, such as
    ; F9=1680k with 21 sectors per trak or 1743k with 21 s/t and 82 tracks,
    ; F0=2880+k ED 2 240 5760 36+ 3.5 80+ etc. F8 harddrive, 0 none, etc.
    BPBSectsPerFAT16   RESW   1  ;Number of Sectors per FAT (0 if FAT32)
    ;DOS v3+ below this line
    BPBSectsPerTrack   RESW   1  ;Number of Sectors per Track
    BPBNumHeads        RESW   1  ;Number of Heads
    BPBHiddenSects     RESD   1  ;Number of Hidden Sectors (DOS 4+)
                                 ;For DOS 3-, this is a Word
                                 ;This is the Number of Sectors on the
                                 ;  Physical Disk that Precede the Partition
                                 ;  Boot Record (including the Master Boot
                                 ;  Record and all Partition Records)
    ;DOS 3 is 11 Bytes Reserved after this
    ;DOS v4+ below this line (Extended BIOS Signature)
    BPBNumSects32      RESD   1  ;Number of Sectors (if DOS 4+ &
                                 ;  NumSects16 = 0)
   ENDS
  BPBSize EQU BIOSParamBlockStruc_size ;25 bytes

  BPBBasicSize EQU (BPBSize+6) ;Officially, BPB Size is 31 Bytes
                               ;  (6 Bytes Reserved after 25 byte base)
  BPBFAT32Size EQU 85 ;For FAT32 Device Driver calls (Category 48h instead
                      ;  of 08h), BPB is 53 bytes + 32 bytes reserved

  ;----------------------------------------------------------------------------
  ;Flash drives do not always have an MBR.  They are sometimes formatted as
  ;  a "super floppy", with a simple Volume Boot Record (just like a Floppy)
  ;  instead of an MBR.  We cannot simply assume that a non-UFI Device will
  ;  have an MBR.
  ;
  ;We'll also need to be able to parse out all of the different DOS versions
  ;  of PBR's, since we on occasion need to return the Volume Label for a
  ;  Disk.  Before DOS Version 4, Volume Names were not included on the Disks,
  ;  so we'll need to use "NO NAME    " for those.
  ;
  ;Note also that the structure of the BPB is determined by the OS that was
  ;  used to format the disk, not necessarily the OS that is currently in
  ;  use.  Therefore, we must be able to process all different types.
  ;----------------------------------------------------------------------------
  STRUC BPBStrucDOS4
    BPB4Common     RESB BPBSize ;Basic Structure
    BPB4DriveNum   RESB      1  ;Physical Drive Number (usually 80h)
    BPB4Reserved1  RESB      1  ;Reserved (0)
    BPB4Signature  RESB      1  ;Extended Boot Record Signature
                                ;References say this should be 28h (40)
                                ;  or 29h (41).  Otherwise non-DOS, e.g. NT.
    BPB4SerialNum  RESD      1  ;Volume Serial Number
    BPB4VolumeName RESB     11  ;Volume Name, space padded
                                ;  "NO NAME   " if never LABELed
    BPB4FileSystem RESB      8  ;ASCII File System Type, space padded
                                ;  Typically something like "FAT16   "
    BPB4Reserved2  RESB      8  ;Reserved for Future Use (0)
   ENDS
  BPB4Size EQU BPBStrucDOS4_size ;59 Bytes

  ;----------------------------------------------------------------------------
  ;Extended BPB for FAT32 drives
  ;----------------------------------------------------------------------------
  STRUC BPBExtStruc
    BPBXCommon          RESB BPBSize ;Basic Structure
    BPBXSectsPerFAT32   RESD      1  ;Sectors/FAT (if SectorsPerFAT16 = 0)
    BPBXFlags           RESW      1  ;Extended Flags (Bits 8-15 reserved?)
      ; BPBXNoFATMirror    EQU   0080h ;Do not Mirror Active FAT (0-4 res.)
      ; BPBXActiveFATMask  EQU   000Fh ;Active FAT (if no Mirroring, 0-based)
    BPBXVersion         RESW      1  ;File System Version (0=Win95 OSR2)
    BPBXRootCluster     RESD      1  ;Starting Cluster of Root Dir
    BPBXInfoSectNum     RESW      1  ;File System Info Sector Num, FFFFh = None
    BPBXBkupBootSect    RESW      1  ;Backup Boot Sector Number, FFFFh = None
    BPBXReserved1       RESB     12  ;Reserved (0)
    BPBXDriveNum        RESB      1  ;Logical Drive Number
    BPBXReserved2       RESB      1  ;Reserved (0)
    BPBXSignature       RESB      1  ;29h FAT32 Boot Record Signature
    BPBXSerialNum       RESD      1  ;Volume Serial Number
    BPBXVolumeName      RESB     11  ;Volume Name, space padded
                                     ;  "NO NAME   " if never LABELed
    BPBXFileSystem      RESB      8  ;ASCII File System Type, space padded
                                     ;  Normally "FAT32   "
   ENDS
  BPBXSize EQU BPBExtStruc_size ;79 Bytes

  ;----------------------------------------------------------------------------
  ;Drive Parameter Block (DPB).  Defined by DOS.
  ;This particular Structure only applies to DOS 4+?
  ;Some parts of the structure are different in previous DOS versions.
  ;DRVEXCH simply swaps DPB's to swap drive letters.
  ;----------------------------------------------------------------------------
  STRUC DriveParamBlockStruc
    DPBDriveNum         RESB  1  ;Drive Number (0=A, 1=B, ...) -1 = Unused
    DPBUnitNum          RESB  1  ;Unit Number within Device Driver
    DPBBytesPerSect     RESW  1  ;Bytes per Sector
    DPBHiSectInCluster  RESB  1  ;Highest Sector Number within a Cluster
                                 ;  (Sectors Per Cluster - 1)
    DPBSectorShift      RESB  1  ;Shift count to convert Clusters to Sectors
    DPBReservedSects    RESW  1  ;Reserved Sectors at beginning of Drive
                                 ;  (Boot Sectors)
    DPBNumFATs          RESB  1  ;Number of FATs (usually 2)
    DPBRootEntries      RESW  1  ;Number of Root Directory Entries
    DPBFirstUserSector  RESW  1  ;First Sector Number containing "real" data
    DPBHighestCluster   RESW  1  ;Highest Cluster Number
                                 ;  (Number of Data Clusters + 1)
                                 ;  If <= 0FF6h, FAT12, else FAT16
    DPBSectorsPerFAT    RESW  1  ;Number of Sectors per FAT
                                 ;  Prior to DOS 4, this was a Byte
    DPBRootSector       RESW  1  ;Sector Number of First (Root) Directory
    DPBDvcDriverPtr     RESD  1  ;Device Driver Header Pointer
    DPBMediaID          RESB  1  ;Media ID
    DPBAccessed         RESB  1  ;Has the Drive been accessed?
      AccessedYes        EQU  0  ;The Drive HAS been Accessed
      AccessedNo         EQU -1  ;The Drive has NOT been accessed
    DPBNextDPBPtr       RESD  1  ;Pointer to next Drive Parameter Block
                                 ;  (-1 = end of list)
    DPBStartCluster     RESW  1  ;Cluster to start search for free space when
                                 ;  writing (usually last cluster allocated)
                                 ;  (for DOS 3+)
                                 ;In DOS 2, = Start Cluster of Current
                                 ;  Directory (0 = Root, -1 = Unknown)
    DPBFreeClustersL    RESW  1  ;Number of Free Clusters left (-1 = Unknown)
                                 ;  For DOS 3+
                                 ;In DOS 2, is 64-byte ASCIIZ Current Directory
  ;Below here is Extended DPB (FAT32 drives only MS-DOS & FreeDOS)
    DPBFreeClustersH    RESW  1  ;Number of Free Clusters left (-1 = Unknown)
                                 ;  High Word
    DPBFATMirror        RESW  1  ;Active FAT/Mirroring
      DPBFATFATNum       EQU 0Fh ;0-based FAT number of active FAT
                                 ;  (only valid if mirroring is disabled)
      ;bits 4-6 = Reserved (0)
      DPBFATNoMirror     EQU 80h ;Do not Mirror Active FAT to Inactive FATs
      ;bits 8-15 = Reserved (0?)
    DPBFSysInfoSect     RESW  1  ;File System Information Sector (-1 = None)
    DPBBkupBootSect     RESW  1  ;Backup Boot Sector (-1 = None)
    DPBFirstClusterSect RESD  1  ;First Sector Number of First Cluster
    DPBMaxCluster       RESD  1  ;Maximum Cluster Number
    DPBNumFATSectors    RESD  1  ;Number of Sectors occupied by FAT
    DPBRootCluster      RESD  1  ;Cluster number of start of Root Directory
    DPBSearchCluster    RESD  1  ;Cluster to start searching for free space
    DPBFiller           RESB  1  ;To align at even addresses
   ENDS
  DPBSize EQU DriveParamBlockStruc_size ;62 bytes (with extra Filler Byte)


;##############################################################################
;TSR DATA
;##############################################################################

;==============================================================================
;Drive Related
;==============================================================================

  ;----------------------------------------------------------------------------
  ;Fixed Disk Parameter Table for a 10MB HD (default to use when no
  ;  Drive is installed yet)
  ;This is copied from the BIOS of a computer with no second disk installed
  ;----------------------------------------------------------------------------
  EVEN 2
  FDPT10MB:
    DW 0132h ;Tracks = 306
    DB   04h ;Heads = 4
    DW 0000h ;Reduced Write Current Cylinder = 0
    DW 0080h ;Precompensation Cylinder = 128
    DB   00h ;Max ECC Burst Length = 0
    DB   00h ;Control Byte = 0
    DB   00h ;Standard Timeout = 0
    DB   00h ;Formatting Timeout = 0
    DB   00h ;Check Drive Timeout = 0
    DW 0131h ;Landing Zone Track = 305
    DB   11h ;Sectors per Track = 17
    DB   00h ;PAD

  ;----------------------------------------------------------------------------
  ;Enhanced Disk Parameter Table for a Drive Larger than 8GB
  ;This is copied from the BIOS of a computer with a Large Hard Drive
  ;----------------------------------------------------------------------------
  EDPTLarge:
    DW 0400h ;Logical Tracks = 1024
    DB  0FFh ;Logical Heads = 255
    DB  0A0h ;Signature for EDPT
    DB   3Fh ;Physical Sectors = 63
    DW 0000h ;Precompensation Cylinder = 0
    DB   00h ;PAD
    DB   08h ;Control Byte = 8
    DW 3FFFh ;Physical Tracks = 16383
    DB   10h ;Physical Heads = 16
    DW 3FFFh ;Landing Zone Track = 16383
    DB   3Fh ;Logical Sectors per Track = 63
    DB   4Bh ;Checksum

  ;----------------------------------------------------------------------------
  ;Data needed while parsing through Partition & Boot Records when a New
  ;  Device is attached
  ;----------------------------------------------------------------------------
  EVEN 2
  NewPartIndex   DB ? ;Current Partition Index
  NewPartOS      DB ? ;Partition Operating System
  NewPartSector  DD ? ;Current Partition Sector
  NewBootSector  DD ? ;Sector number of Volume Boot Sector
  NewSectsInPart DD ? ;Number of Sectors in Partition

  LBAInBuffer    DQ ? ;LBA of the Sector that is in the Sector Buffer
  DiskInBuffer   DB ? ;Disk Number of the Sector that is in the Sector Buffer

  ;----------------------------------------------------------------------------
  ;Our Partition Stack
  ;Keeps track of Extended Partitions as we process our way through all
  ;  of the Partition Entries on a Drive.
  ;----------------------------------------------------------------------------
  PStkEntries EQU 16 ;Maximum Partition Levels we can process
  EVEN 2
  PartStack:
    TIMES (PStkEntries*PStkSize) DB 0

  PartStackPtr DW PartStack ;Pointer to Current Partition Stack Entry

  ;----------------------------------------------------------------------------
  ;Valid Partition Operating System Flags that we (DOS) can handle
  ;Does this need to change based on DOS version?
  ;----------------------------------------------------------------------------
  ValidPOSList:
    DB POSFAT12      ;01h FAT-12
    DB POSFAT16      ;04h Fat-16 (<= 32M)
    DB POSDOSExt     ;05h DOS Extended
    DB POSFAT16H     ;06h Fat-16 (> 32M)
    DB POSFAT32      ;0Bh Fat-32 (<= 8G?)
    DB POSFAT32LBA   ;0Ch Fat-32 w/ LBA Extensions (<= 2T)
    DB POSLBAVFAT16H ;0Eh FAT16H w/ LBA Extensions
    DB POSLBAVFATExt ;0Fh DOSExt w/ LBA Extensions
  ;  DB POSFAT32H     ;1Bh FAT-32 Hidden
  ;  DB POSFAT32HH    ;1Ch FAT-32 Hidden w/ LBA Extension
  ;  DB POSLBAVFATH   ;1Eh LBA VFAT Hidden
  ;  DB POSDOS33Sec   ;F2h DOS 3.3+ Secondary
    DB POSGPT        ;GUID Partition Table
    DB 0 ;End of Table


;==============================================================================
;GPT / GUID Related
;==============================================================================

  ;----------------------------------------------------------------------------
  ;GUID Signature for a Microsoft Basic Data Disk.
  ;This is the GPT entry for all FAT-related GPT partitions, in addition
  ;  to several other types, including NTFS.
  ;----------------------------------------------------------------------------
  EVEN 2
  MSBasicDataGUID:
    DB 0A2h,0A0h,0D0h,0EBh,0E5h,0B9h,033h,044h,
    DB 087h,0C0h,068h,0B6h,0B7h,026h,099h,0C7h

  ;----------------------------------------------------------------------------
  ;Temporary variables needed while processing the GPT entries in
  ;  the background.
  ;----------------------------------------------------------------------------
  GPTTotalEntries   DW  ? ;Number of GPT Table Entries Left to Process
  GPTCurrentIndex   DW  ? ;Current GPT Index Number we're processing
                          ;  (GPT Table Index)
  GPTTblSector      DW  ? ;Sector that GPT Table Entry is in (2+)
  GPTLastPartSector DQ  ? ;Last Sector of Current Partition
  GPTTblOffset      DW  ? ;Offset in GPT Sector of Current Entry
  GPTEntrySize      DW  ? ;Size of each GPT Table Entry
  GPTCurrentSubCtr  DB  ? ;Current GPT Loop Counter we're processing
                          ;  (GPT Entries within a Single Sector)
  GPTEntriesPerSect DB  ? ;Number of GPT Table Entries per Sector
                          ;  (usually 4)
  GPTWriteProt      DB No ;GPT Entry marked as Read-Only
  GPTDriveHandled   DB No ;Drive has been handled/processed



;##############################################################################
;CODE TO DECODE THE PARTITION RECORDS ON A DISK
;##############################################################################

;------------------------------------------------------------------------------
;Inputs: DS = EMS/DPMS Segment/Selector
;        ES = TSR Segment/Selector
;        BX = Disk Info Entry
;        GS:[SectorBuffer]
;        Partition Stack contains Partition Details
;NOTES: If we are told to mount non-FAT drives, we only do some simple
;         tests to verify that the disk is actually mountable.
;       All we do is test the size, and if it's less than approximately
;         4M sectors (2GB capacity with "normal" 512-byte sectors)
;         we will always mount it.  If its between approximatley 4M
;         and 4G sectors (2TB capacity with "normal" 512-byte sectors)
;         we only mount it if the OS supports FAT32.  Not all versions
;         of DOS can actually support 2GB disks (e.g. DOS 3 only supports
;         32 MB disks), though we don't test or account for that possibility
;         here.  We may need to change this someday.
;       At least in theory, the only times you would want to mount a non-FAT
;         drive would be to format it, or if you had some sort of special
;         driver installed for the specific format type on the partition.
;         I'm not aware of any such driver that exists today, though
;         somebody may eventually make one for NTFS or some of the ext?
;         *nix partitions.  Even if such a driver is written,
;         however, it will probably use raw (INT 13h) functions to access
;         the disk rather than depending on DOS mounting things itself.
;         It is much more likely that mounting non-FAT partitions would
;         only be used temporarily to format something and then the option
;         would be turned off.
;------------------------------------------------------------------------------

GotPartSectorCode:
  CALL SetupRETFRegs2        ;DS:BX = Disk, DS:DI = Device,
                             ;ES:SI = Int14Request, AX = Error
                             ;GS = TSR Data Segment/Selector
  ; MOV  B [NewDvcStage],190   ;Update Stage (status, for debugging only)
  JNZ .P70                   ;If error, quit
  MOV  B [BX+DskDrvTimer],0  ;Mark as Media Inserted
  ; MOV  B [NewDvcStage],191   ;Update Stage (status, for debugging only)
  CMP  D [LBAInBuffer],0     ;Are we handling Sector 0?
  JNE .P03                   ;If not, continue
.P01:                        ;We're handling Sector 0
;  TEST B [BX+DskFlags3],(DskFlagNoFormat+DskFlagNonFAT)
;                             ;Unformatted or Non-FAT CD/DVD?
;  JNZ .P10                   ;If so, treat as a SuperFloppy
  CMP  W [BX+DskOtherOffset],0 ;CD/DVD?
  JNE .P10                   ;If so, handle as superfloppy
  CALL TestSectBuffSame      ;Any data in Sector 0?
  JNZ .P03                   ;If so, continue
  OR   B [BX+DskFlags3],DskFlagNoFormat ;If not, mark as unformatted
  JMP .P10                   ;Treat as a (Super)Floppy

.P03:                        ;Done with Sector 0 stuff
  CMP  W [GS:SectorBuffer+MBRSignature],BootRecSig ;Valid MBR/Partition Record?
  JNE .P70                   ;If not, error
.P05:                        ;Valid MBR/Partition Record
  ; MOV  B [NewDvcStage],192   ;Update Stage (status, for debugging only)
  CALL CopyBPBData           ;Copy BPB data, if appropriate
  JC  .P70                   ;If error, quit
  CALL TestForFloppy         ;Is it a (Super)Floppy?
  JC  .P15                   ;If not, handle a hard drive

.P10:                        ;Is a (Super)Floppy
  CALL InitNewFloppy         ;Initialize Drive Parameters
  JMP GotBootSectorCode      ; continue with boot sector
  ; MOV  AX,GotBootSectorXX    ;Where to go
  ; MOV  BP,GotBootSectorPM    ;  when done
  ; MOV  W [NextSCSIReqOfst],GotBootSectorCode ; go here when call is done
  ; MOV  AX,GotBootSectorCode
  ; CALL Start1TimeIntSI       ;Issue a 1-time interrupt (do a 1 ms delay)
  ; JNC .P14                   ;If it worked, we're done (jump to a RETF)
  ; CALL FinishUpBadDrive      ;Else: Error, mark it as Bad, then quit
  ; .P14
  ; RETF

.P15:                        ;Not a (Super)Floppy
  ; MOV  B [NewDvcStage],193   ;Update Stage (status, for debugging only)
  MOV  SI,[PartStackPtr]     ;[SI] = Partition Stack Pointer
  MOV  BP,SI                 ;[BP] = Partition Stack Pointer
  MOV  SI,[SI+PStkOffset]    ;[SI] = Partition Entry Offset
  CMP  SI,(SectorBuffer+MBRPart4) ;Processed all Partitions?
  JA  .P50                   ;If so, handle it
.P20:                        ;Loop to here to find valid Partitions
  MOV  AL,[GS:SI+PartOpSys]  ;Get the Operating System Flag
  OR   AL,AL                 ;Valid Partition Record?
  JZ  .P40                   ;If not, skip it
  CALL TestValidPOS          ;If so, is it an OS that we can handle?
;  JNC .P23                   ;If so, continue
;  TEST B [EMSFlags],MountNonFAT ;If not, should we mount it anyway?
;  JZ  .P40                   ;If not, skip it
;  CMP  D [GS:SI+PartLBALength],MaxFAT16Sects ;Small enough to be FAT16?
;  JBE .P25                   ;If so, handle it
;  TEST B [EMSFlags],UseFAT32 ;If not, can we mount FAT32 partitions?
;  JNZ .P25                   ;If so, handle it
;  JMP .P40                   ;If not, skip it
  JC  .P40                   ;If not, skip it
.P23:                        ;Partition OS is OK
  CMP  AL,POSDOSExt          ;If so, is it an Extended DOS Partition?
  JE  .P30                   ;If so, handle it
  CMP  AL,POSLBAVFATExt      ;If not, is it an Extended W95 Partition?
  JE  .P30                   ;If so, handle it
  CMP  AL,POSGPT             ;GUID Partition Table?
  JNE .P25                   ;If not, continue
  CALL TestGPTDisk           ;If so, verify it's really a GPT disk
  JC  .P70                   ;If not, Error
.P24:                        ;GPT, not MBR
  ; MOV  B [NewDvcStage],194   ;Update Stage (status, for debugging only)
  ; MOV  W [NextSCSIReqOfstRM],StartGPTXX ;Go here
  ; MOV  W [NextSCSIReqOfstPM],StartGPTPM ;  when done
  MOV  W [NextSCSIReqOfst],StartGPTCode ; go here when SCSI I/O complete
  MOV  EAX,1                 ;Read Sector 1
;  XOR  EDX,EDX               ;  Sector 1
;  CALL SCSIReadSectorBG      ;Start the Read Sector Request
  CALL SCSIReadSectorEAX0BG ;Issue the Read Sector Request for sector EAX
  JC  .P70                   ;If error, quit
  CALL SetupSCSIDone         ;Issue the Request
  JC  .P70                   ;If error, handle it
  JMP .P90                   ;If OK, we're done

.P25:                        ;Is a regular DOS Partition
  ; MOV  B [NewDvcStage],195   ;Update Stage (status, for debugging only)
  ADD  W [DS:BP+PStkOffset],PRecordSize ;Update Partition Stack pointer
  CALL InitNewHard           ;Initialize Drive Parameters (returns EAX)
  ; MOV  W [NextSCSIReqOfstRM],GotBootSectorXX ;Go here
  ; MOV  W [NextSCSIReqOfstPM],GotBootSectorPM ;  when done
  MOV  W [NextSCSIReqOfst],GotBootSectorCode ; go here when SCSI I/O complete
  CALL SCSIReadSectorEAXBG   ;Start Read Sector Request for sector EAX (?)
  JC  .P70                   ;If error, quit
  CALL SetupSCSIDone         ;Issue the Request
  JC  .P70                   ;If error, handle it
  JMP .P90                   ;Done

.P30:                        ;Is an extended DOS partition
  ; MOV  B [NewDvcStage],196   ;Update Stage (status, for debugging only)
  CALL PushPart              ;Push the Current Partition Info (EAX)
  JNC .P60                   ;If it worked, continue
  JMP .P70                   ;If error, quit

.P40:                        ;Not a valid Partition
  ADD  SI,PRecordSize        ;Point at the next Partition Entry
  MOV  [DS:BP+PStkOffset],SI ;Store it for next time
  CMP  SI,(SectorBuffer+MBRPart4) ;Processed all Partitions?
  JBE .P20                   ;If not, keep looking
.P50:                        ;Have processed all Partitions in Record
  CALL PopPart               ;Go to the previous Partition (CF, EAX)
  JC  .P80                   ;If none, we've processed all of them
.P60:                        ;Get the next Partition Record
  ; MOV  W [NextSCSIReqOfstRM],GotPartSectorXX ;Go here
  ; MOV  W [NextSCSIReqOfstPM],GotPartSectorPM ;  when done
  MOV  W [NextSCSIReqOfst],GotPartSectorCode ; go here when SCSI I/O complete
  CALL SCSIReadSectorEAXBG   ;Start Read Sector Request for sector EAX (?)
  JC  .P70                   ;If error, quit
  CALL SetupSCSIDone         ;Issue the Request
  JNC .P90                   ;If it worked, we're done
  ; MOV  B [NewDvcStage],197   ;Update Stage (status, for debugging only)
.P70:                        ;Error
  CALL FinishUpBadDrive      ;Mark it as Bad
  JMP .P90                   ;Done

.P80:                        ;Completely done
  AND  B [EMSFlags],~(Int14ReqInUse+SectBuffInUse) ;Mark resources as available
.P90:                        ;Done
  RETF

;------------------------------------------------------------------------------
;Inputs: DS = EMS/DPMS Segment/Selector
;        ES = TSR Segment/Selector
;        BX = Disk Info Entry
;        GS:[SectorBuffer]
;NOTES: The FAT32 mods (changing DPBRootEntries & DPBSectorsPerFAT)
;         is a patch needed for FreeDOS.  MS-DOS does not need the patch.
;       If the disk/partition appears to be formatted (has some sort of
;         "random" data in it), we check for the boot record signature word
;         at offset 510.  If the signature is missing, we won't assign
;         a Drive Letter.  We will assign a Drive Letter if it appears
;         to be Unformatted, or contains the proper boot record signature.
;------------------------------------------------------------------------------

GotBootSectorCode:
  CALL SetupRETFRegs2               ;DS:BX = Disk, DS:DI = Device,
                                    ;ES:SI = Int14Request, AX = Error
                                    ;GS = TSR Data Segment/Selector
  ; MOV  B [NewDvcStage],200          ;Update Stage (status, for debugging only)
  JNZ .B70                          ;If error, quit
  ; MOV  B [NewDvcStage],201          ;Update Stage (status, for debugging only)
  CMP  W [BX+DskOtherOffset],0      ;CD/DVD (unformatted/non-FAT)?
  JNE .B20                          ;If so, assign a drive letter
;  TEST B [BX+DskFlags3],(DskFlagNoFormat+DskFlagNonFAT)
;                                    ;Unformatted Disk/CD/DVD or non-FAT CD/DVD?
;  JZ  .B20                          ;If so, assign a Drive Letter
  CALL TestSectBuffSame             ;Is the boot sector formatted?
  JNZ .B20                          ;If so, assign a Drive Letter
  CMP  W [GS:SectorBuffer+MBRSignature],BootRecSig ;Valid VBR/PBR Sector?
  JNE .B70                          ;If not, error
.B20:                               ;Need to assign a Drive Letter
  ; MOV  B [NewDvcStage],202          ;Update Stage (status, for debugging only)
  CALL GetAvailDriveSI              ;Get an available Drive Entry ([SI])
  JC  .B85                          ;If error, we're done
  MOV  [SI+DrvDskOffset],BX         ;If OK, store the Disk Offset
  CALL CopyNewDriveParams           ;Copy Parameters to Drive Entry
  TEST B [EMSFlags],IsFreeDOS       ;Is this FreeDOS?
  JZ  .B40                          ;If not, continue
  CMP  B [SI+DrvPartOS],POSFAT32    ;Is it FAT32?
  JE  .B30                          ;If so, handle it
  CMP  B [SI+DrvPartOS],POSFAT32LBA ;Is it FAT32?
  JNE .B40                          ;If not, no mods needed
.B30:                               ;Is FAT32 under FreeDOS
  CALL BPB2DPBFreeDOS32             ;Partially fill in the DPB
.B40:                               ;Done FreeDOS with FAT32 mods
  OR   B [SI+DrvFlags],DrvFlagReady ;Mark Drive as configured
  MOV  B [SI+DrvGIOAccess],1        ;Mark as Accessible via Generic IOCTL
  CMP  B [SI+DrvPartOS],-1          ;Is it a (Super)Floppy?
  JE  .B80                          ;If so, we're done
  MOV  EAX,[SI+DrvPartSector]       ;EAX = Next Partition Sector to read
  ; MOV  W [NextSCSIReqOfstRM],GotPartSectorXX ;Go here
  ; MOV  W [NextSCSIReqOfstPM],GotPartSectorPM ;  when done
  MOV  W [NextSCSIReqOfst],GotPartSectorCode ; go here when SCSI I/O complete
  CALL SCSIReadSectorEAXBG          ;Start Read Sector Request for sector EAX (?)
  JC  .B70                          ;If error, quit
  CALL SetupSCSIDone                ;Issue the Request
  JNC .B90                          ;If it worked, we're done
  ; MOV  B [NewDvcStage],203          ;Update Stage (status, for debugging only)
.B70:                               ;Error
  CALL FinishUpBadDrive             ;Mark it as Bad
  JMP .B90                          ;Done

.B80:                               ;Completely done
  MOV  B [BX+DskDrvTimer],0         ;Mark as Media Inserted
.B85:                               ;Done
  AND  B [EMSFlags],~(Int14ReqInUse+SectBuffInUse) ;Mark resources available
.B90:                               ;Done
  ; CALL WriteStage                 ; Show current NewDvcStage value (to debug)
  RETF


;##############################################################################
;CODE TO PROCESS A GPT (INSTEAD OF MBR) DISK
;##############################################################################

;------------------------------------------------------------------------------
;START DECODING A GPT DISK FOR DRIVE ENTRIES
;Inputs: DS = EMS/DPMS Segment/Selector
;        ES = TSR Segment/Selector
;        BX = Disk Info Entry
;        GS:[SectorBuffer] (contains data from LBA Sector 1)
;NOTES: GPT has a second copy of the boot sector information stored at the
;         end (last several sectors) of the disk.  We don't bother looking
;         at the end of the disk, either for verification that the data
;         is valid or for the actual data.  We do some validity tests
;         (such as signatures and UUID/GUID's). but for the most part
;         just assume that the data at the beginning of the disk is correct.
;------------------------------------------------------------------------------

StartGPTCode:
  CALL SetupRETFRegs2                   ;DS:BX = Disk, DS:DI = Device,
                                        ;ES:SI = Int14Request, AX = Error
                                        ;GS = TSR Data Segment/Selector
  ; MOV  B [NewDvcStage],220              ;Update Stage (status, for debugging only)
  MOV  BP,SectorBuffer                  ;GS:BP = Sector Buffer
  CMP  D [GS:BP+GPTHSignature],'EFI '   ;Correct Signature?
  JNE .U70                              ;If not, Error
  CMP  D [GS:BP+GPTHSignature+4],'PART' ;Correct Signature?
  JNE .U70                              ;If not, Error
  CMP  D [GS:BP+GPTHPAD1],0             ;Correct PAD?
  JNE .U70                              ;If not, Error
  CMP  D [GS:BP+GPTHThisHdrLBA+4],0     ;Correct LBA?
  JNE .U70                              ;If not, Error
  CMP  D [GS:BP+GPTHThisHdrLBA],1       ;Correct LBA?
  JNE .U70                              ;If not, Error
  CMP  D [GS:BP+GPTHPartTblLBA+4],0     ;Correct LBA?
  JNE .U70                              ;If not, Error
  CMP  D [GS:BP+GPTHPartTblLBA],2       ;Correct LBA?
  JNE .U70                              ;If not, Error
  MOV  EAX,D [GS:BP+GPTHOtherHdrLBA+4]  ;Is the High Word of the
  CMP  EAX,D [BX+DskMaxLBA+4]           ;  Other Header LBA too big?
  JA  .U70                              ;If so, error
  JB  .U20                              ;If no need to test Low Word, skip
  MOV  EAX,D [GS:BP+GPTHOtherHdrLBA]    ;Is the Low Word of the
  CMP  EAX,D [BX+DskMaxLBA]             ;  Other Header LBA too big?
  JA  .U70                              ;If so, error
.U20:                                   ;Other Header LBA is OK
  MOV  EAX,[GS:BP+GPTHPartEntries]      ;EAX = Number of GPT Table Entries
  TEST EAX,0FFFF_0000h                  ;Is it reasonable?
  JNZ .U70                              ;If not, error
  CMP  EAX,128                          ;Is it valid?
  JB  .U70                              ;If not, Error
  MOV  [GPTTotalEntries],AX             ;If OK, Store it
  MOV  ECX,[GS:BP+GPTHPartEntrySz]      ;ECX = Size of Each GPT Table Entry
  TEST ECX,0FFFF_0000h                  ;Is it reasonable?
  JNZ .U70                              ;If not, error
  CMP  CX,128                           ;Is it valid?
  JB  .U70                              ;If not, Error
  MOV  AX,[BX+DskBytesPerSect]          ;Is it more than
  CMP  CX,AX                            ;  one Sector?
  JA  .U70                              ;If so, error
  MOV  [GPTEntrySize],CX                ;If OK, store it
  XOR  DX,DX                            ;Divide Bytes per Sector by
  DIV  CX                               ;  Bytes per GPT Table Entry
  MOV  [GPTEntriesPerSect],AL           ;Store GPT Table Entries per Sector
  MOV  B [GPTCurrentSubCtr],0           ;Store as Starting Loop Counter
  MOV  W [GPTCurrentIndex],0            ;Current Index = 0
  MOV  W [GPTTblOffset],SectorBuffer    ;Store Starting GPT Sector Offset
  MOV  B [GPTDriveHandled],No           ;Mark as unprocessed
  ; MOV  W [NextSCSIReqOfstRM],NextGPTXX  ;Go here
  ; MOV  W [NextSCSIReqOfstPM],NextGPTPM  ;  when done
  MOV  W [NextSCSIReqOfst],NextGPTCode ; go here when SCSI I/O complete
  MOV  EAX,2                            ;Get LBA Sector 2
  ; MOV  B [NewDvcStage],221              ;Update Stage (status, for debugging only)
  CALL SCSIReadSectorEAX0BG             ;Start Read Sector Request for sector EAX
  JC  .U70                              ;If error, quit
  CALL SetupSCSIDone                    ;Issue the Request
  JNC .U90                              ;If no Error, quit
  ; MOV  B [NewDvcStage],222              ;Update Stage (status, for debugging only)
.U70:                                   ;Error
  CALL FinishUpBadDrive                 ;Mark it as Bad
.U90:                                   ;Done
  RETF

;------------------------------------------------------------------------------
;DECODE THE NEXT SECTOR OF GPT TABLE ENTRIES
;Inputs: DS = EMS/DPMS Segment/Selector
;        ES = TSR Segment/Selector
;        BX = Disk Info Entry
;        [LBAInBuffer]
;        [GPTTotalEntries]
;        [GPTCurrentIndex]
;        [GPTTblSector]
;        [GPTEntrySize]
;        [GPTCurrentSubCtr]
;        [GPTEntriesPerSect]
;        [GPTDriveHandled]
;        TSRDataSegment:[SectorBuffer] (contains GPT Sector Data)
;        CLD already issued
;------------------------------------------------------------------------------

NextGPTCode:
  CALL SetupRETFRegs2             ;DS:BX = Disk, DS:DI = Device,
                                  ;ES:SI = Int14Request, AX = Error
                                  ;GS = TSR Data Segment/Selector
  MOV  AX,W [LBAInBuffer]         ;Store the Current
  MOV  [GPTTblSector],AX          ;  Sector Number
  MOV  BP,[GPTTblOffset]          ;GS:BP = Sector (GPT) Data
  CMP  B [GPTDriveHandled],Yes    ;Returning from a processed Drive?
  MOV  B [GPTDriveHandled],No     ;Mark as unprocessed for next time
  JE  .X50                        ;If so, don't process it again
.X10:                             ;Loop to here for each GPT Entry
  PUSH DI,ES                      ;Save Sector Buffer Pointer
  MOV  ES,GS                      ;ES:DI =
  MOV  DI,BP                      ;  String to Test
  MOV  CX,4                       ;Test 4 Words (entire 16-byte GUID)
  MOV  SI,MSBasicDataGUID         ;DS:SI = String to try and match
  REPE CMPSD                      ;Compare the Strings
  POP  ES,DI                      ;Restore Sector Buffer Pointer
  JNZ .X50                        ;If no match, skip it
.X20:                             ;GPT has correct GUID
  MOV  EDX,[GS:BP+GPTELastLBA+4]  ;EDX:EAX =
  MOV  EAX,[GS:BP+GPTELastLBA]    ;  Number
  SUB  EAX,[GS:BP+GPTEFirstLBA]   ;  of
  SBB  EDX,[GS:BP+GPTEFirstLBA+4] ;  Sectors
  ADD  EAX,1                      ;  in
  ADC  EDX,0                      ;  Partition
  JNZ .X50                        ;If > 4G Sectors, we can't use it in DOS
.X30:                             ;GUID & Size are OK
  MOV  AL,[GS:BP+GPTEAttribs+7]   ;AL = High Byte of GPT Entry Attributes
  TEST AL,0C0h                    ;Hidden or Do-Not-Assign?
  JNZ .X50                        ;If so, skip it
  MOV  B [GPTWriteProt],No        ;Assume not Read-only
  TEST AL,10h                     ;Is it Read-Only?
  JNZ .X34                        ;If so, handle it
  MOV  AL,[GS:BP+GPTEAttribs]     ;AL = Low Byte of GPT Entry Attributes
  TEST AL,01h                     ;System (Do-Not Change) Partition?
  JZ  .X35                        ;If not, continue
.X34:                             ;Read-only
  MOV  B [GPTWriteProt],Yes       ;Mark as Read-Only
.X35:                             ;Read-only Status is stored
  MOV  EAX,[GS:BP+GPTELastLBA]    ;Store
  MOV  [GPTLastPartSector],EAX    ;  Last
  MOV  EAX,[GS:BP+GPTELastLBA+4]  ;  Sector of
  MOV  [GPTLastPartSector+4],EAX  ;  Partition
  ; MOV  W [NextSCSIReqOfstRM],GotGPTSectorXX ;Go here
  ; MOV  W [NextSCSIReqOfstPM],GotGPTSectorPM ;  when done
  MOV  W [NextSCSIReqOfst],GotGPTSectorCode ; go here when SCSI I/O complete
  MOV  EDX,[GS:BP+GPTEFirstLBA+4] ;EDX:EAX =
  MOV  EAX,[GS:BP+GPTEFirstLBA]   ;  Partition Boot Sector (VBR)
  JMP .X65                        ;Do it

.X50:                             ;Not a GPT Entry we can use
  ADD  BP,[GPTEntrySize]          ;GS:BP = Next GPT Entry
  MOV  AX,[GPTCurrentIndex]       ;Increment
  INC  AX                         ;  Current Index
  CMP  AX,[GPTTotalEntries]       ;Done with the entire GPT Table yet?
  MOV  [GPTCurrentIndex],AX       ;Store Incremented Index
  JAE .X80                        ;If so, we're done
  MOV  AL,[GPTCurrentSubCtr]      ;If not,
  INC  AL                         ;  Increment Sector Loop Counter
  CMP  AL,[GPTEntriesPerSect]     ;Done with this Sector yet?
  MOV  [GPTCurrentSubCtr],AL      ;Store Updated Counter
  JB  .X10                        ;If not, continue
.X60:                             ;Done with this Sector
  MOV  B [GPTCurrentSubCtr],0     ;Reset Sector Index Counter
  ; MOV  W [NextSCSIReqOfstRM],NextGPTXX ;Go here
  ; MOV  W [NextSCSIReqOfstPM],NextGPTPM ;  when done
  MOV  W [NextSCSIReqOfst],NextGPTCode ; go here when SCSI I/O complete
  MOVZX EAX,W [GPTTblSector]      ;Get
  INC  EAX                        ;  Next
  XOR  EDX,EDX                    ;  LBA Sector
.X65:                             ;EAX = Next Sector to Download
  CALL SCSIReadSectorBG           ;Start Read Sector Request for sector EDX:EAX
  JC  .X70                        ;If error, quit
  CALL SetupSCSIDone              ;Issue the Request
  JNC .X90                        ;If OK, we're done
.X70:                             ;Error
  CALL FinishUpBadDrive           ;Handle Bad Drive
  JMP .X90                        ;Done

.X80:                             ;OK
  MOV  B [BX+DskDrvTimer],0       ;Mark as Media Inserted
  AND  B [EMSFlags],~(Int14ReqInUse+SectBuffInUse) ;Mark resources available
.X90:                             ;Done
  RETF

;------------------------------------------------------------------------------
;DECODE THE NEXT SECTOR OF GPT TABLE ENTRIES
;Inputs: DS = EMS/DPMS Segment/Selector
;        ES = TSR Segment/Selector
;        BX = Disk Info Entry
;        [LBAInBuffer]
;        [GPTTotalEntries]
;        [GPTCurrentIndex]
;        [GPTTblSector]
;        [GPTEntrySize]
;        [GPTCurrentSubCtr]
;        [GPTEntriesPerSect]
;        [GPTLastPartSector]
;        [GPTDriveHandled]
;        [GPTWriteProt]
;        TSRDataSegment:[SectorBuffer] (contains Boot Sector Data)
;        CLD already issued
;------------------------------------------------------------------------------

GotGPTSectorCode:
  CALL SetupRETFRegs2                   ;DS:BX = Disk, DS:DI = Device,
                                        ;ES:SI = Int14Request, AX = Error
                                        ;GS = TSR Data Segment/Selector
  MOV  BP,SectorBuffer                  ;GS:BP = Sector Buffer
  TEST B [EMSFlags],MountNonFAT         ;Do we care how it's formatted?
  JNZ .S30                              ;If not, go ahead and mount it
  CALL TestSectBuffSame                 ;Does the buffer contain any data?
  JZ  .S30                              ;If not, it's unformatted, so mount it
.S10:                                   ;Boot Sector contains data
  CALL GetVolNamePtr                    ;Get the Volume Name Pointer
                                        ;  TSRDataSegment:[SI] -- just used
                                        ;  to test for valid VBR, we don't
                                        ;  actually care what's returned in SI
  JC  .S90                              ;If error, quit
  MOV  AL,[GS:BP+11+BPBXSignature]      ;Is it
  CMP  AL,29h                           ;  FAT32?
  JNE .S30                              ;If not, continue
  TEST B [EMSFlags],UseFAT32            ;If so, can we Mount it?
  JZ  .S90                              ;If not, we're done
.S30:                                   ;Sector is unformatted, formatted
                                        ;  appropriately (FAT), or we don't
                                        ;  care if/how it's formatted
  CALL GetAvailDriveSI                  ;Get an available Drive Entry (DS:SI)
  JC  .S90                              ;If error, we're done
  MOV  [SI+DrvDskOffset],BX             ;If OK, store Disk Offset
  CALL CopyGPTDriveParams               ;Copy Parameters to Drive Entry
  TEST B [EMSFlags],IsFreeDOS           ;Is this FreeDOS?
  JZ  .S40                              ;If not, continue
  CMP  AL,29h                           ;If so, is the Drive FAT32?
  JNE .S40                              ;If not, continue
  CALL BPB2DPBFreeDOS32                 ;IF so, partially fill in the DPB
.S40:                                   ;Done FreeDOS with FAT32 mods
  OR   B [SI+DrvFlags],DrvFlagReady     ;Mark Drive as configured
  MOV  B [SI+DrvGIOAccess],1            ;Mark as Accessible via Generic IOCTL
  MOV  B [BX+DskDrvTimer],0             ;Mark as Media Inserted
  CMP  B [GPTWriteProt],Yes             ;Is it Read-only?
  JNE .S90                              ;If not, continue
  OR   B [SI+DrvFlags],DrvFlagWriteProt ;If so, mark it as Read-only
.S90:                                   ;Done with this Drive
  MOV  B [GPTDriveHandled],Yes          ;Mark as processed
  ; MOV  W [NextSCSIReqOfstRM],NextGPTXX  ;Go here
  ; MOV  W [NextSCSIReqOfstPM],NextGPTPM  ;  when done
  MOV  W [NextSCSIReqOfst],NextGPTCode ; go here when SCSI I/O complete
  MOVZX EAX,W [GPTTblSector]            ;EAX = Sector to Download
  CALL SCSIReadSectorEAX0BG             ;Start Read Sector Request for sector EAX
                                        ;  (Just Assume It Works)
  CALL SetupSCSIDone                    ;Issue the Request
                                        ;  (Just Assume It Works)
  RETF

;------------------------------------------------------------------------------
;TEST A GPT BOOT SECTOR FOR VALIDITY
;Inputs:  DS    = EMS/DPMS Segment/Selector
;         ES    = TSR Segment/Selector
;         GS    = TSR Data Segment/Selector
;         DS:BX = Disk Info Entry
;         DS:DI = Disk Info Entry
;         DS:SI = Offset of Partition Entry
;         DS:[LBAInBuffer]
;         GS:[SectorBuffer] filled with Sector Data
;Outputs: CF = Clear if OK (GPT Record Appears Valid)
;              [BX].DskFlags2
;            = Set if Error (Invalid GPT Data)
;NOTE: This simply tests the data that is in Sector 0 of the Disk.  There are
;        several more tests that could be made to verify validity, but
;        they require downloading other sectors, so we will handle that
;        later (in LookForNewDrives), not here (in LookForNewDevices).
;      In hybrid MBR/GPT schemes, the LBA value in the MBR does not cover
;        the entire disk -- it just covers the GPT partition.  There are
;        then dual entries for the "extra" MBR partitions: an entry in the
;        MBR, and a corrseponding entry in the GPT.
;      There actually is no "standard" way to do hybrid MBR/GPT schemes,
;        so we will just use the GPT.  Note that Windows does not do this --
;        if the system is a hybrid, Windows uses the MBR instead of the GPT.
;        AFAIK, MS is the only OS vendor that does this.  Everybody else
;        uses the GPT in the case of a hybrid scheme.
;      Windows also has a required "MS Reserved" partition that must be on
;        the disk, and requires that it be located before the Basic Data
;        partitions (which are the ones we care about).  We don't care
;        whether there is an MS Reserved Partition or not, or whether it
;        comes first, or anything at all like that.  All we care about is
;        whether a partition is designated as MS Basic Data or not, if/how
;        it is formatted, and the GPT flags related to hidden/read-only/ignore.
;      One of our original tests included verifying that the MBR entry for
;        the GPT partition included the entire disk, but this is not true
;        in hybrid systems, so is not a valid test.  You gotta love MS
;        screwing with the standards.
;------------------------------------------------------------------------------
TestGPTDisk:
  PUSH EDI                         ;Save used registers
  CMP  D [LBAInBuffer],0           ;Is the Data in the Buffer from Sector 0?
  JNE .G70                         ;If not, Error
  CMP  SI,(SectorBuffer+MBRPart1)  ;Is it the first Partition?
  JNE .G70                         ;If not, Error
  MOV  DI,(SectorBuffer+MBRPart1)  ;GS:[DI] = MBR Partition Record
  CMP  D [GS:DI+PartLBASector],1   ;Correct Starting LBA?
  JE  .G80                         ;If so, it's OK
.G70:                              ;Error
  STC                              ;Set return flag
  JMP .G90                         ;Done

.G80:                              ;OK
  OR   B [BX+DskFlags2],DskFlagGPT ;Mark as a GPT Disk
  CLC                              ;Set Return Flag
.G90:                              ;Done
  POP  EDI                         ;Restore used registers
  RET

;------------------------------------------------------------------------------
;TEST THE SECTOR BUFFER TO SEE IF IT CONTAINS A SINGLE, REPEATED BYTE VALUE
;Inputs:   DS = EMS Segment/Selector
;          ES = TSR Segment/Selector
;          BX = Disk Info Entry
;         [BX].DskBytePerSector = Number of bytes to test
;         TSRDataSegment:[SectorBuffer] contains data
;         CLD already issued
;Outputs: ZF = Set if buffer contains all the same value
;            = Clear if not
;------------------------------------------------------------------------------
TestSectBuffSame:
  PUSH AX,CX,DI,ES     ;Save used registers
  CALL GetTSRDataSegES ;ES:DI =
  MOV  DI,SectorBuffer ;  Sector Buffer
  MOV  CX,[BX+DskBytesPerSect] ;CX = Number of Bytes to Test
  MOV  AL,[ES:DI]      ;AL = Value to look for
  REPE SCASB           ;Test the Buffer
  POP  ES,DI,CX,AX     ;Restore used registers
  RET


;##############################################################################
;SUPPORT CODE TO LOOK FOR NEW DRIVES
;##############################################################################

;------------------------------------------------------------------------------
;COPY BPB DATA TO DISK INFO ENTRY, IF APPROPRIATE
;Inputs:  DS = EMS/DPMS Segment/Selector
;         ES = TSR Segment/Selector
;         BX = Disk Info Entry (may already be filled in)
;         [LBAInBuffer]
;         GS:[SectorBuffer]
;Outputs: CF = Clear if OK
;              Copies & verifies MaxLBA, CHS values if not already filled in
;            = Set if Error
;NOTES: When we get to here, GS:[SectorBuffer] contains data from one of the
;         disk sectors.  Depending on how the disk responded to the
;         earlier SCSI requests (Format Capacity, Capacity, and
;         Flex Descriptor) we may or may not know anything about the
;         disk size parameters (Max LBA or CHS values).  If we got valid
;         data from the earlier requests, we just ignore what we find in
;         Sector 0.  If we didn't get any data from the earlier requests,
;         we use what we find in Sector 0 to "guesstimate" the disk
;         parameters.  This may not necessarily be accurate, depending on
;         how (or if) the disk was formatted.
;------------------------------------------------------------------------------
CopyBPBData:
  PUSH EAX,CX,DX,DI,SI,BP        ;Save used registers
  CMP  W [BX+DskSectsPerTrack],0 ;Already have valid data?
  JNE .B80                       ;If so, we're done
  CMP  D [LBAInBuffer],0         ;Is this Sector 0?
  JNE .B80                       ;If not, just quit
  MOV  SI,SectorBuffer           ;GS:SI = Sector Buffer
  CMP  W [GS:SI+MBRSignature],BootRecSig ;Valid MBR/Partition Record?
  JNE .B70                       ;If not, error
  MOV  AL,[SI]                   ;Get the first OpCode Byte in the Sector
  CMP  AL,OpCodeJmpShort         ;Is it a Boot Record?
  JE  .B20                       ;If so, it's a (Super)Floppy
  CMP  AL,OpCodeJmpNear          ;Is it a Boot Record?
  JE  .B20                       ;If so, it's a (Super)Floppy
.B10:                            ;It's a Hard Drive
  CALL GetBPBHard                ;Glean Hard Drive Data (EAX, CX, DX, DI, BP)
  JZ  .B70                       ;If error, quit
  JMP .B30                       ;If OK, continue

.B20:                            ;It's a (Super)Floppy
  CALL GetBPBFloppy              ;Glean Floppy Data  (EAX, CX, DX, DI, BP)
  JZ  .B70                       ;If error, quit
.B30:                            ;Handle Max LBA (EAX)
  MOV  D [BX+DskMaxLBA+4],0      ;Store
  MOV  [BX+DskMaxLBA],EAX        ;  Max LBA
  MOV  [BX+DskBytesPerSect],CX   ;Store Bytes per Sector
  MOV  [BX+DskNumTracks],DX      ;Store Number of Tracks
  MOV  [BX+DskNumHeads],DI       ;Store Number of Heads
  MOV  [BX+DskSectsPerTrack],BP  ;Store Sectors per Track
.B60:                            ;All values stored
  JMP .B80                       ;Done

.B70:                            ;Error
  STC                            ;Set error flag
  JMP .B90                       ;Done

.B80:                            ;OK
  CALL AdjustCHSValues           ;Adjust CHS values as necessary
  CALL FillInFDPTParams          ;Fill in FDPT Table if needed
  CLC                            ;Set return flag
.B90:                            ;Done
  POP  BP,SI,DI,DX,CX,EAX        ;Restore used registers
  RET

;------------------------------------------------------------------------------
;GLEAN THE DISK SIZE INFORMATION FROM A HARD DRIVE SECTOR 0 PARTITION RECORD
;Inputs:  DS    = EMS/DPMS Segment/Selector
;         ES    = TSR Segment/Selector
;         GS:SI = Sector Buffer (with Sector 0 Partition Record)
;Outputs: ZF  = Clear if OK (data is valid)
;               EAX  = Max LBA (Total number of Sectors - 1)
;                CX  = Number of Bytes per Sector
;                DX  = Number of Cylinders/Tracks
;                DI  = Number of Heads
;                BP  = Number of Sectors per Track
;         ZF  = Set if Error
;               EAX, CL, DX, BP = Undefined
;NOTES: Normally, partition records are set up in ascending order from the
;         beginning of the disk (record 1 is nearest the beginning of the
;         disk, and record 4 is nearest the end).  Howevever, this is not an
;         absolute requirement and we will not assume that the disk is set
;         up that way.  We will check all of the records and find the one
;         with the largest ending LBA.
;       Also, normally, partitions are set up so that they start at the
;         beginning of a cylinder (head 0, sector 1), and stop at the end of
;         a cylinder (max cylinder, max sector).  However, this is not an
;         absolute requirement, either.  If it is not, though, we will never
;         be able to figure out what the maximum head and sector numbers are.
;         We will search all of the records and find the one with the
;         largest numbers, though we would expect all of the records to
;         be the same.  Most disks only have one partition anyway, so we
;         won't have anything to compare the first record to.  As long as
;         the disk was formatted in a "reasonable" customary fashion, we should
;         be OK.
;       Partition records also do not contain any information at all about
;         the number of bytes per sector, so we will simply assume it is 512.
;         That is a fairly safe assumption, though it is certainly not
;         ALWAYS true.
;------------------------------------------------------------------------------
GetBPBHard:
  PUSH EBX,SI,DS              ;Save used registers
  MOV  DS,GS                  ;DS:SI = Sector Buffer
  MOV  BP,1                   ;Initialize Sectors per Track
  XOR  DI,DI                  ;Initialize Number of Heads
  XOR  DX,DX                  ;Initialize Number of Cylinders/Tracks
  XOR  EBX,EBX                ;EBX = LBA comparison value
  MOV  CX,4                   ;Do 4 partition records
  ADD  SI,MBRPart1            ;[SI] = First Partition Record
.H10:                         ;Loop to here for each partition record
  MOV  EAX,[SI+PartLBASector] ;Calculate
  ADD  EAX,[SI+PartLBALength] ;  End LBA + 1
  OR   EAX,EAX                ;Valid record?
  JZ  .H60                    ;If not, skip it
.H20:                         ;Valid record
  CMP  EAX,EBX                ;Is it more than the previous record(s)?
  JBE .H30                    ;If not, it's not the last record
  MOV  EBX,EAX                ;If so, mark it as the new last record
.H30:                         ;Get Max Cylinder
  MOV  AL,[SI+PartStartCyl]   ;AX =
  MOV  AH,[SI+PartStartSect]  ;  Start Cylinder
  SHR  AH,6                   ;  Number
  CMP  AX,DX                  ;Is it more than the previous Max Cylinder?
  JBE .H35                    ;If not, continue
  MOV  DX,AX                  ;If so, store it as the new Max Cylinder
.H35:                         ;DX = Updated Max Cylinder
  MOV  AL,[SI+PartEndCyl]     ;AX =
  MOV  AH,[SI+PartEndSect]    ;  End Cylinder
  SHR  AH,6                   ;  Number
  CMP  AX,DX                  ;Is it more than the previous Max Cylinder?
  JBE .H40                    ;If not, continue
  MOV  DX,AX                  ;If so, store it as the new Max Cylinder
.H40:                         ;Get Max Head
  XOR  AH,AH                  ;Values from here on are bytes
  MOV  AL,[SI+PartStartHead]  ;AX = Start Head Number
  CMP  AX,DI                  ;Is it more than the previous Max Head?
  JBE .H45                    ;If not, continue
  MOV  DI,AX                  ;If so, store it as the new Max Head
.H45:                         ;DI = Updated Max Head
  MOV  AL,[SI+PartEndHead]    ;AX = End Head Number
  CMP  AX,DI                  ;Is it more than the previous Max Head?
  JBE .H50                    ;If not, continue
  MOV  DI,AX                  ;If so, store it as the new Max Head
.H50:                         ;Get Max Sector
  MOV  AL,[SI+PartStartSect]  ;AX =
  AND  AL,PSectSector         ;  Start Sector Number
  CMP  AX,BP                  ;Is it more than the previous Max Cylinder?
  JBE .H55                    ;If not, continue
  MOV  BP,AX                  ;If so, store it as the new Max Cylinder
.H55:                         ;DX = Updated Max Cylinder
  MOV  AL,[SI+PartEndSect]    ;AX =
  AND  AL,PSectSector         ;  End Sector Number
  CMP  AX,BP                  ;Is it more than the previous Max Cylinder?
  JBE .H60                    ;If not, continue
  MOV  BP,AX                  ;If so, store it as the new Max Cylinder
.H60:                         ;Done with this Partition Record
  ADD  SI,PRecordSize         ;Point at the next record
  LOOP .H10                   ;Keep going until we're done
  MOV  EAX,EBX                ;EAX =
  DEC  EAX                    ;  Max LBA
  INC  DX                     ;Convert Max Cylinder to Number of Cylinders
  INC  DI                     ;Convert Max Head to Number of Heads
  MOV  CX,512                 ;CX = Bytes per Sector (assume 512)
  CMP  BP,1                   ;Set return flag
  POP  DS,SI,EBX              ;Restore used registers
  RET

;------------------------------------------------------------------------------
;GLEAN THE DISK SIZE INFORMATION FROM A (SUPER)FLOPPY SECTOR 0 BOOT RECORD
;Inputs:  DS    = EMS/DPMS Segment/Selector
;         ES    = TSR Segment/Selector
;         GS:SI = Sector Buffer (with Sector 0 Floppy Boot Record)
;Outputs: ZF  = Clear if OK (data is valid)
;               EAX  = Max LBA (Total number of Sectors - 1)
;                CX  = Number of Bytes per Sector
;                DX  = Number of Cylinders/Tracks
;                DI  = Number of Heads
;                BP  = Number of Sectors per Track
;         ZF  = Set if Error
;               EAX, CL, DX, BP = Undefined
;------------------------------------------------------------------------------
GetBPBFloppy:
  PUSH SI,DS                      ;Save used registers
  MOV  DS,GS                      ;DS:SI =
  ADD  SI,11                      ;  BIOS Parameter Block
  XOR  BP,BP                      ;Set test value (Sectors per Track) to 0
  MOV  CX,[SI+BPBBytesPerSect]    ;Get the Bytes per sector
  CMP  CX,[DOSBytesPerSect]       ;Is it too big for DOS to handle?
  JA  .F90                        ;If so, Error
  MOV  DI,[SI+BPBNumHeads]        ;DI = Number of Heads
  MOV  AX,[SI+BPBSectsPerTrack]   ;BP =
  MUL  DI                         ;  Number of Heads times
  MOV  BP,AX                      ;  Number of Sectors Per Track
  XOR  DX,DX                      ;DX:AX = Total Number
  MOV  AX,[SI+BPBNumSects16]      ;  of Sectors
  OR   AX,AX                      ;Is the Total a 16-bit value?
  JNZ .F20                        ;If so, continue
.F10:                             ;Total Sectors is a 32-bit value
  MOV  AX,[SI+BPBNumSects32]      ;DX:AX = Total Number
  MOV  DX,[SI+BPBNumSects32+2]    ;  of Sectors
.F20:                             ;DX:AX = Total Number of Sectors
  DIV  BP                         ;AX = Whole Cylinders
  OR   DX,DX                      ;Do we need to round it up?
  JZ  .F30                        ;If not, continue
  INC  AX                         ;If so, round it up
.F30:                             ;AX = Number of Cylinders
  MOV  DX,AX                      ;DX = Number of Cylinders
  MOVZX EAX,W [SI+BPBNumSects16]  ;EAX = 16-bit Number of Sectors
  OR   EAX,EAX                    ;Is it a 16-bit LBA?
  JNZ .F40                        ;If so, continue
  MOV  EAX,[SI+BPBNumSects32]     ;Get the Number of Sectors
.F40:                             ;EAX = Number of Sectors
  MOVZX ECX,W [SI+BPBHiddenSects] ;Add in the number of
  ADD  EAX,ECX                    ;  Hidden Sectors
  DEC  EAX                        ;Convert to Max LBA
  MOV  BP,[SI+BPBSectsPerTrack]   ;BP = Sectors Per Track
.F90:                             ;Done
  OR   BP,BP                      ;Set return flag
  POP  DS,SI                      ;Restore used registers
  RET

;------------------------------------------------------------------------------
;TEST THE CYLINDER/HEAD/SECTOR VALUES WE'VE GOT SO FAR FOR LEGITIMACY
;Inputs:  DS = EMS/DPMS Segment/Selector
;         ES = TSR Segment/Selector
;         BX = Disk Info Entry
;                Cylinder, Head, Sector Values filled in (as downloaded
;                 via SCSI Requests)
;Outputs:  Cylinder, Head, Sector Values adjusted as necessary to make
;            them work with INT 13h Functions
;NOTES: Several USB Drives I've tested don't return CHS values that
;         correspond to the Total Sectors value.  This makes the adjustments
;         necessary so that everything "fits" like it's supposed to.
;       If the CHS values are "legal" and match up with the Total Sectors,
;         we leave them alone.  If the Total Sectors is small enough to be
;         accessible using INT 13h Functions (<= 8.4GB), this adjusts
;         the CHS values so that they work correctly.  If the drive is too
;         big for INT 13h functions, this sets the size to the maximum (8.4GB).
;       Some Devices return Total Number of Sectors when they should return
;         Maximum LBA.  Since disks should always have an even number of
;         sectors (Maximum LBA should always be odd) we'll perform the
;         adjustment here to make sure the Maximum LBA is odd.
;------------------------------------------------------------------------------
AdjustCHSValues:
  PUSH EAX,CX,DX,EBP                     ;Save used registers
  TEST B [BX+DskMaxLBA],1                ;Is the Max LBA odd?
  JNZ .A05                               ;If so, it's OK
  SUB  D [BX+DskMaxLBA],1                ;If not,
  SBB  D [BX+DskMaxLBA+4],0              ;  make it odd
.A05:                                    ;MaxLBA adjusted, if appropriate
  CMP  D [BX+DskMaxLBA+4],0              ;MaxLBA too big?
  JNE .A10                               ;If so, handle it
  MOV  EBP,[BX+DskMaxLBA]                ;EBP =
  INC  EBP                               ;  Total Sectors
  JZ  .A10                               ;If overflow, it's too big
  CMP  D [BX+DskMaxLBA],MaxI13TotalSects ;Can the Disk use INT 13h funcs?
  JBE .A20                               ;If so, handle it
.A10:                                    ;Disk is too big for INT 13h funcs
  MOV  W [BX+DskNumHeads],MaxI13Heads    ;Store Max Heads
  MOV  W [BX+DskNumTracks],MaxI13Tracks  ;Store Max Cylinders
  MOV  W [BX+DskSectsPerTrack],MaxI13SectsPerTrack ;Store Max Sects Per Track
  JMP .A90                               ;Done

.A20:                                    ;Disk can use INT 13h funcs
  CMP  W [BX+DskNumHeads],0              ;Did we have any CHS data at all?
  JE  .A40                               ;If not, adjustments are needed
  CMP  W [BX+DskNumHeads],MaxI13Heads    ;Is Number of Heads Legal?
  JA  .A40                               ;If not, adjustments are needed
  CMP  W [BX+DskNumTracks],MaxI13Tracks  ;Is Number of Tracks Legal?
  JA  .A40                               ;If not, adjustments are needed
  CMP  W [BX+DskSectsPerTrack],MaxI13SectsPerTrack  ;Is Sectors Per Track Legal?
  JA  .A40                               ;If not, adjustments are needed
  MOV  AX,[BX+DskNumHeads]               ;Multiply Number of Heads
  MUL  W [BX+DskSectsPerTrack]           ;  by Number of Sectors per Track
  MOV  CX,AX                             ;Save this intermediate value
  MUL  W [BX+DskNumTracks]               ;Multiply by Number of Tracks
  PUSH DX                                ;Put Total
  PUSH AX                                ;  Sectors in
  POP  EAX                               ;  EAX
  CMP  EAX,EBP                           ;How does it compare to Total Sectors?
  JE  .A90                               ;If equal, it's OK
  JA  .A40                               ;If too big, adjustments are needed
  MOV  AX,[BX+DskNumTracks]              ;If too small,
  INC  AX                                ;  is it
  MUL  CX                                ;  too small
  PUSH DX                                ;  by
  PUSH AX                                ;  exactly
  POP  EAX                               ;  one
  CMP  EAX,EBP                           ;  cylinder?
  JA  .A90                               ;If so, it's OK
.A40:                                    ;Adjustments are needed to CHS values
  MOV  W [BX+DskNumHeads],MaxI13Heads    ;Store Max Heads
  MOV  W [BX+DskSectsPerTrack],MaxI13SectsPerTrack ;Store Max Sects Per Track
  MOV  AX,[BX+DskMaxLBA]                 ;Calculate the
  MOV  DX,[BX+DskMaxLBA+2]               ;  Number of Tracks
  MOV  CX,(MaxI13Heads*MaxI13SectsPerTrack) ;  we need
  DIV  CX                                   ;  to use
  MOV  [BX+DskNumTracks],AX              ;Store it
.A90:                                    ;Done
  POP  EBP,DX,CX,EAX                     ;Restore used registers
  RET

;------------------------------------------------------------------------------
;COPY THE APPROPRIATE PARAMETERS FROM A DISK INFO ENTRY TO A FDPT
;Inputs:  DS = EMS/DPMS Segment/Selector
;         ES = TSR Segment/Selector
;         BX = Disk Info Entry to use
;Outputs: Copies Appropriate Data from Disk Info Entry to FBPT
;------------------------------------------------------------------------------
FillInFDPTParams:
  PUSH AX,DI                        ;Save used registers
  MOV  AL,[BX+DskInt13DrvNum]       ;AL = Int 13h Drive / CD Unit
  MOV  DI,FDPT80h                   ;Assume first hard drive
  CMP  AL,80h                       ;First hard drive?
  JE  .F10                          ;If so, continue
  MOV  DI,FDPT81h                   ;If not, assume second hard drive
  CMP  AL,81h                       ;Second hard drive?
  JNE .F90                          ;If not, quit
.F10:                               ;[BX] = Disk Info, [DI] = FDPT
  CALL EmptyFDPTTable               ;Fill Table with Zeroes
  CMP  D [BX+DskMaxLBA],MaxI13TotalSects ;Is it a large disk?
  JBE .F30                          ;If not, continue
.F20:                               ;Is a Large Disk
  CALL CopyEDPTLarge                ;  with the Large Data
  JMP .F90                          ;Done

.F30:                               ;Not a large disk
  CALL EmptyFDPTTable               ;Fill the table with zeroes
  MOV  AX,[BX+DskNumTracks]         ;Copy the
  MOV  [ES:DI+FDPTNumTracks],AX     ;  number of Tracks
  MOV  [ES:DI+FDPTLandingZone],AX   ;Store in Landing Zone also
  MOV  AL,[BX+DskSectsPerTrack]     ;Copy the
  MOV  [ES:DI+FDPTSectsPerTrack],AL ;  Sectors per Track
  MOV  AX,[BX+DskNumHeads]          ;Copy the
  MOV  [ES:DI+FDPTNumHeads],AL      ;  number of Heads
  CMP  AL,8                         ;More than 8 heads?
  JBE .F90                          ;If not, we're done
  OR   B [ES:DI+FDPTCtlByte],FDPTHeads8Plus ;If so, set Flag in Control Byte
.F90:                               ;Done
  POP  DI,AX                        ;Restore used registers
  RET

;------------------------------------------------------------------------------
;COPY THE DEFAULT OR LARGE FDPT/EPDT PARAMETERS TO AN FDPT TABLE
;Inputs:  DS    = EMS/DPMS Segment/Selector
;         ES    = TSR Segment/Selector
;         ES:DI = FDPT Table to fill in
;         CLD already issued
;Outputs: Fills in FBPT with appropriate data
;------------------------------------------------------------------------------
CopyFDPT10MB:
  PUSH SI           ;Save used registers
  MOV  SI,FDPT10MB  ;[SI] = 10MB FDPT Table Data
  JMP  CopyXDPT     ;Do it

CopyEDPTLarge:
  PUSH SI           ;Save used registers
  MOV  SI,EDPTLarge ;[SI] = Large FDPT Table Data
;  JMP  CopyXDPT     ;Do it
CopyXDPT:
  PUSH CX,DI ;Save used registers
  MOV  CX,(FDPTSize/2) ;CX = Number of Words to Copy
  REP  MOVSW ;Copy the Data
  POP  DI,CX ;Restore used registers
  POP  SI    ;Restore used registers
  RET

;------------------------------------------------------------------------------
;CREATE A PARTIALLY CORRECT DPB GIVEN A BPB
;Inputs:  DS = EMS/DPMS Segment/Selector
;         ES = TSR Segment/Selector
;         SI = Drive Info Entry (BPB filled in)
;Outputs: DPB area filled in
;NOTE: This is only called if in FreeDOS and drive is FAT32 (this patches a
;        bug in FreeDOS -- the problem is not actually in USBDRIVE).
;        Newer versions of FreeDOS will hopefully have the bug fixed.
;      This does not actually create a real, valid DPB.  It just copies the
;        things from the BPB that are straightforward to copy (no calculations
;        needed), and sets everything else to 0 or -1.  This is apparently
;        close enough to a real DPB that it keeps the bug in FreeDOS happy.
;------------------------------------------------------------------------------
BPB2DPBFreeDOS32:
  PUSH EAX,BX,ECX,EDX,DI,ES            ;Save used registers
  XOR  ECX,ECX                         ;ECX = 0
  MOV  EDX,-1                          ;EDX = -1
  LEA  BX,[SI+DrvBPB]                  ;DS:BX = BPB
  CALL GetTSRVarSegES                  ;ES:DI =
  MOV  DI,[SI+DrvDPBOffset]            ;  DPB
  MOV  AX,[BX+BPBBytesPerSect]         ;Copy
  MOV  [ES:DI+DPBBytesPerSect],AX      ;   Bytes per Sector
  MOV  AL,[BX+BPBSectsPerCluster]      ;Copy
  DEC  AL                              ;  Sectors
  MOV  [ES:DI+DPBHiSectInCluster],AL   ;  per Cluster
  XOR  AL,AL                           ;Start Sector->Cluster Shift at 0
  MOV  AH,[BX+BPBSectsPerCluster]      ;AH = Sectors Per Cluster
  OR   AH,AH                           ;Is it valid?
  JZ  .B15                             ;If not, leave it at 0
.B10:                                  ;Loop here to find Sector->Cluster Shift
  SHR  AH,1                            ;Shift over by 1
  JC  .B15                             ;If CF, this is bit we're looking for
  INC  AL                              ;If not, increment Sector Shift
  JMP .B10                             ;Keep looking

.B15:                                  ;AL = Sector Shift
  MOV  [ES:DI+DPBSectorShift],AL       ;Store Sector Shift
  MOV  EAX,[BX+BPBReservedSects]       ;Copy
  MOV  [ES:DI+DPBReservedSects],EAX    ;  Reserved Sectors,
  MOV  AL,[BX+BPBRootEntries+1]        ;  Number of FATs,
  MOV  [ES:DI+DPBRootEntries+1],AL     ;  and Number of Root Entries
  MOV  [ES:DI+DPBFirstUserSector],DX   ;First User Sector = Unknown
  MOV  [ES:DI+DPBHighestCluster],CX    ;Highest Cluster = 0
  MOV  AX,[BX+BPBSectsPerFAT16]        ;Copy
  MOV  [ES:DI+DPBSectorsPerFAT],AX     ;  Sectors per FAT
  MOV  [ES:DI+DPBRootSector],DX        ;Root Sector = -1
  ;DPBDvcDriverPtr filled in elsewhere
  MOV  AL,[BX+BPBMediaID]              ;Copy
  MOV  [ES:DI+DPBMediaID],AL           ;  Media ID
  ;DPBAccessed filled in elsewhere
  ;DPBNextDPBPtr filled in elsewhere
  MOV  [ES:DI+DPBStartCluster],DX      ;Search Start Cluster = Unknown
  MOV  [ES:DI+DPBFreeClustersL],EDX    ;Number of Free Clusters = Unknown
  MOV  AX,[BX+BPBXFlags]               ;Copy
  MOV  [ES:DI+DPBFATMirror],AX         ;  Extended Flags
  MOV  EAX,[BX+BPBXInfoSectNum]        ;Copy File System Information Sector
  MOV  [ES:DI+DPBFSysInfoSect],EAX     ;  & Backup Boot Sector
  MOV  [ES:DI+DPBFirstClusterSect],EDX ;First Cluster Sector = Unknown
  MOV  [ES:DI+DPBMaxCluster],EDX       ;Maximum Cluster = Unknown
  MOV  [ES:DI+DPBNumFATSectors],EDX    ;Number of FAT Sectors = Unknown
  MOV  EAX,[BX+BPBXRootCluster]        ;Copy
  MOV  [ES:DI+DPBRootCluster],EAX      ;  Root Cluster
  MOV  [ES:DI+DPBSearchCluster],ECX    ;Search Cluster = Unknown
  POP  ES,DI,EDX,ECX,BX,EAX            ;Restore used registers
  RET

;------------------------------------------------------------------------------
;RESET AN FDPT TABLE TO ALL ZEROES
;Inputs:  DS    = EMS/DPMS Segment/Selector
;         ES    = TSR Segment/Selector
;         ES:DI = FDPT Table to fill in
;         CLD already issued
;Outputs: Fills in FBPT with zeroes
;------------------------------------------------------------------------------
EmptyFDPTTable:
  PUSH AX,CX,DI ;Save used registers
  XOR  AX,AX    ;Fill the table with zeroes
  MOV  CX,(FDPTSize/2) ;CX = Number of Words to Copy
  REP  STOSW    ;Fill the Table
  POP  DI,CX,AX ;Restore used registers
  RET

;------------------------------------------------------------------------------
;TEST A PARTITION RECORD TO SEE IF THE DRIVE IS A (SUPER)FLOPPY OR NOT
;Inputs:  DS = EMS/DPMS Segment/Selector
;         ES = TSR Segment/Selector
;         BX = Disk Info Entry
;         GS:[SectorBuffer]
;         [LBAInBuffer] = LBA of Sector data that is in the Buffer
;         Partition Stack contains Partition Details
;         [SectorBuffer] filled with Data from Partition Sector
;Outputs: CF = Clear if this is the MBR for a (Super)Floppy
;            = Set if this is Data from a Partition Record
;NOTES: We check here for Volume Boot Records (Floppies & Super-Floppies),
;         Master Boot Records (hard drives), and Multi-boot Records
;         (hard drives that look like Super-Floppies).  We do a few
;         relatively simple tests to distinguish between VBR's and
;         MBR's, but it's possible that we could be fooled by some exotic
;         IPL without more extensive testing.  It is UNLIKELY that we would
;         be fooled, but it is possible.
;------------------------------------------------------------------------------
TestForFloppy:
  PUSH EAX,BX,CX,SI,DS            ;Save used registers
  CMP  D [LBAInBuffer],0          ;Is the Data in the Buffer from Sector 0?
  JNE .F70                        ;If not, it can't be (Super)Floppy
  MOV  SI,[PartStackPtr]          ;[SI] =
  MOV  SI,[SI+PStkOffset]         ;  Partition Stack Offset
  CMP  SI,(SectorBuffer+MBRPart1) ;Are we processing the first record?
  JNE .F70                        ;If not, it can't be a (Super)Floppy
  MOV  DS,GS                      ;DS:SI =
  MOV  SI,SectorBuffer            ;  Sector Buffer
  CMP  B [SI],OpCodeJmpShort      ;Is it a possible Volume Boot Record?
  JE  .F10                        ;If so, keep checking
  CMP  B [SI],OpCodeJmpNear       ;Is it a possible Volume Boot Record?
  JNE .F70                        ;If not, it can't be a (Super)Floppy
.F10:                             ;Is possibly a (Super)Floppy
  MOV  BL,2                       ;BL = Active Partition Counter
  ADD  SI,MBRPart1                ;[SI] = Partition 1 Record
  CMP  B [SI+PartOpSys],POSNone   ;Is it a valid OpSys type?
  JE  .F80                        ;If not, it's a (Super)Floppy
  MOV  CX,4                       ;Need to test a total of 4 partition records
.F20:                             ;Loop to here for each partition
  MOV  AL,[SI+PartPrimary]        ;AL = Primary (Active) Partition Indicator
  MOV  AH,[SI+PartOpSys]          ;AH = OpSys type
  OR   AX,AX                      ;Is it an Invalid OpSys on an Inactive Partition?
  JNZ .F25                        ;If not, continue testing
  CMP  D [SI+PartLBASector],0     ;If so, is the Start LBA Sector 0?
  JNE .F80                        ;If not, it's a (Super)Floppy
  CMP  D [SI+PartLBALength],0     ;Is the Partition Length 0?
  JE  .F50                        ;If so, it's probably a valid Partition Record
.F25:                             ;Not an Invalid OpSys on and Inactive Partition
  OR   AL,AL                      ;Inactive Partition?
  JZ  .F30                        ;If so, check OpSys
  CMP  AL,80h                     ;Active Partition?
  JNE .F80                        ;If so, it's a (Super)Floppy
  DEC  BL                         ;Decrement Active Partition Counter
  JZ  .F80                        ;If >1 Active Partition, it's a (Super)Floppy
.F30:                             ;Valid Primary Flag
  OR   AH,AH                      ;Valid OpSys?
  JZ  .F80                        ;If not, it's a (Super)Floppy
.F40:                             ;Valid Partition Flag & OpSys
  MOV  EAX,[SI+PartLBASector]     ;Get the Starting LBA Sector
  OR   EAX,EAX                    ;Is it 0?
  JZ  .F80                        ;If so, it's a (Super)Floppy
  CMP  D [SI+PartLBALength],0     ;Is the Partition Length 0?
  JE  .F80                        ;If so, it's a (Super)Floppy
  ADD  EAX,[SI+PartLBALength]     ;Add the Length to the Start
  JC  .F80                        ;If overflow, it's a (Super)Floppy
.F50:                             ;Partition record is in a valid format
  ADD  SI,PRecordSize             ;Point at the next partition record
  LOOP .F20                       ;Keep testing
.F70:                             ;Not a (Super)Floppy
  STC                             ;Set return flag
  JMP .F90                        ;Done

.F80:                             ;Is a (Super)Floppy
  CLC                             ;Set return flags
.F90:                             ;Done
  POP  DS,SI,CX,BX,EAX            ;Restore used registers
  RET

;------------------------------------------------------------------------------
;INITIALIZE THE PARTITION STACK TO HANDLE A NEW DISK
;Inputs:  DS  = EMS/DPMS Segment/Selector
;         ES  = TSR Segment/Selector
;Outputs: [PartStackPtr] = First Partition Stack Entry
;         First Partition Stack Entry initialized to Sector 0, Partition Index 0
;------------------------------------------------------------------------------
InitPartStack:
  PUSH BX                  ;Save used registers
  MOV  BX,PartStack        ;DS:BX = First (Current) Partition Stack Record
  MOV  [PartStackPtr],BX   ;Store it as the Current Pointer
  MOV  D [BX+PStkSector],0 ;First Sector LBA is always 0
  MOV  W [BX+PStkOffset],(SectorBuffer+MBRPart1) ;First Index = 0
  POP  BX                  ;Restore used registers
  RET

;------------------------------------------------------------------------------
;PUSH A PARTITION ENTRY ON TO THE PARTITION STACK
;Inputs   DS = EMS/DPMS Segment/Selector
;         ES = TSR Segment/Selector
;         GS:[SectorBuffer]
;         SI = Current Partition Entry (in SectorBuffer)
;                Points at Record for Extended Partition!
;         [PartStackPtr]
;Outputs: CF = Clear if OK (Space left in Stack)
;              [PartStackPtr] incremented appropriately
;              Next Partition Details stored on Partition Stack
;              EAX = Sector Number of Next Partition to Retrieve
;            = Set if Error
;              EAX unchanged
;------------------------------------------------------------------------------
PushPart:
  PUSH BX                            ;Save used registers
  MOV  BX,[PartStackPtr]             ;[BX] = Current Partition Stack Pointer
  CMP  BX,PartStack+((PStkEntries-1)*PStkSize) ;Any room left on the Stack?
  JAE .H70                           ;If not, error
  MOV  [BX+PStkOffset],SI            ;Put Next Partition Record Offset
  ADD  W [BX+PStkOffset],PRecordSize ;  on the Stack
  ADD  BX,PStkSize                   ;If OK, update the Stack Pointer
  MOV  [PartStackPtr],BX             ;Store it
  MOV  W [BX+PStkOffset],(SectorBuffer+MBRPart1) ;Start at Index 0
  MOV  EAX,[GS:SI+PartLBASector]     ;Get the Partition LBA Sector
  CMP  BX,(PartStack+(PStkSize*2))   ;Are we in an Extended Partition?
  JB  .H20                           ;If not, it's OK like it is
  ADD  EAX,[PartStack+PStkSize+PStkSector] ;If so, the Sector is Relative to
                                           ;  the first Extended Partition
.H20:                                ;Stack entries are set up correctly
  MOV  [BX+PStkSector],EAX           ;Store the Absolute Partition Sector
  JMP .H80                           ;Done

.H70:                                ;Error
  STC                                ;Set return flag
  JMP .H90                           ;Done

.H80:                                ;OK
  CLC                                ;Set return flag
.H90:                                ;Done
  POP  BX                            ;Restore used registers
  RET

;------------------------------------------------------------------------------
;POP A PARTITION ENTRY OFF OF THE PARTITION STACK
;Inputs:  DS = EMS/DPMS Segment/Selector
;         ES = TSR Segment/Selector
;         [PartStackPtr]
;Outputs: CF = Clear if OK (there is something on the Stack)
;              [PartStackPtr] decremented appropriately
;              EAX = Sector Number of Next Partition to Retrieve (Previous Part)
;            = Set if Error
;              EAX = Unchanged
;------------------------------------------------------------------------------
PopPart:
  PUSH BX                  ;Save used registers
  MOV  BX,[PartStackPtr]   ;[BX] = Current Partition Stack Pointer
  CMP  BX,PartStack        ;Already popped everything off the stack?
  JBE .P70                 ;If so, error
  SUB  BX,PStkSize         ;If OK, update the Stack Pointer
  MOV  [PartStackPtr],BX   ;Store it
  MOV  EAX,[BX+PStkSector] ;EAX = Sector Number of Previous Partition
  JMP .P80                 ;Done

.P70:                      ;Error
  STC                      ;Set return flag
  JMP .P90                 ;Done

.P80:                      ;OK
  CLC                      ;Set return flag
.P90:                      ;Done
  POP  BX                  ;Restore used registers
  RET

;------------------------------------------------------------------------------
;CONVERT PARTITION RECORD OFFSET (IN SectorBuffer) TO A PARTITION RECORD INDEX
;Inputs:  SI = Current Partition Record (in SectorBuffer)
;Outputs: CF = Clear if OK (SI is valid)
;              AL = Index Number of Partition Record (0-3)
;            = Set if Error (SI too big/too small)
;              AL = Unchanged
;------------------------------------------------------------------------------
PartOffset2Idx:
  PUSH BX    ;Save used registers
  CMP  SI,(SectorBuffer+MBRPart1) ;Is SI too small?
  JB  .X70   ;If so, error
  CMP  SI,(SectorBuffer+MBRPart4) ;Is SI too big?
  JA  .X70   ;If so, error
  MOV  BX,SI ;If OK, put Pointer in BX
  SUB  BX,(SectorBuffer+MBRPart1) ;Convert to 0-based offset
  SHR  BX,4  ;Divide by 16 (size of each Partition Record)
  MOV  AL,BL ;Store result in AL
  JMP .X80   ;Done

.X70:        ;Error
  STC        ;Set return flag
  JMP .X90   ;Done

.X80:        ;OK
  CLC        ;Set return flag
.X90:        ;Done
  POP  BX    ;Restore used registers
  RET

;------------------------------------------------------------------------------
;SETUP NEW DRIVE PARAMETERS AS NEEDED FOR A (SUPER) FLOPPY DRIVE
;Inputs:  DS = EMS/DPMS Segment/Selector
;         ES = TSR Segment/Selector
;         BX = Disk Info Entry
;Outputs: [NewPartOS], [NewPartSector], etc. as needed for (Super)Floppy
;------------------------------------------------------------------------------
InitNewFloppy:
  PUSH EAX                  ;Save used registers
  MOV  B [NewPartOS],-1     ;Partition OS = -1 (Flag for Floppy)
  MOV  D [NewPartSector],-1 ;Partition Sector = -1 (N/A)
  MOV  B [NewPartIndex],-1  ;Partition Index = -1 (N/A)
  MOV  D [NewBootSector],0  ;Boot Sector = 0
  MOV  EAX,[BX+DskMaxLBA]   ;Store
  INC  EAX                  ;  Number of
  MOV  [NewSectsInPart],EAX ;  Sectors
  POP  EAX                  ;Restore used registers
  RET

;------------------------------------------------------------------------------
;SETUP NEW DRIVE PARAMETERS AS NEEDED FOR A HARD DRIVE
;Inputs:  DS    = EMS/DPMS Segment/Selector
;         ES    = TSR Segment/Selector
;         GS    = TSR Data Segment
;         GS:SI = Partition Record (in SectorBuffer)
;         [LBAInBuffer] = Sector Number of data that is Stored in SectorBuffer
;Outputs: [NewPartOS], [NewPartSector], etc. as needed for Hard Drive
;         EAX = LBA of Boot Sector to retrieve
;------------------------------------------------------------------------------
InitNewHard:
  MOV  AL,[GS:SI+PartOpSys]      ;Get the Operating System Flag
  MOV  [NewPartOS],AL            ;Store Partition OS
  MOV  EAX,[LBAInBuffer]         ;Store
  MOV  [NewPartSector],EAX       ;  Partition Sector

  MOV  [NewBootSector],EAX       ;Store Absolute Boot Sector Start

  CALL PartOffset2Idx            ;Store
  MOV  [NewPartIndex],AL         ;  Partition Index
  MOV  EAX,[GS:SI+PartLBALength] ;Store the
  MOV  [NewSectsInPart],EAX      ;  Number of Sectors
  MOV  EAX,[GS:SI+PartLBASector] ;Store the
  ADD  [NewBootSector],EAX       ;  Boot Sector
  RET

;------------------------------------------------------------------------------
;TEST A PARTITION OPERATING SYSTEM BYTE TO SEE IF WE SUPPORT IT
;Inputs:  DS = EMS/DPMS Segment/Selector
;         ES = TSR Segment/Selector
;         AL = Partition Operating System Byte to test
;         [EMSFlags].UseFAT32
;         ValidPOSList
;         CLD Already issued
;Outputs: CF = Clear if OK (OS is in the Table)
;            = Set if Error (OS is not in the Table)
;------------------------------------------------------------------------------
TestValidPOS:
  PUSH AX,SI           ;Save used registers
  MOV  SI,ValidPOSList ;[SI] = POS Table
  MOV  AH,AL           ;AH = Byte to Look for
.O10:                  ;Loop to here for each table entry
  LODSB                ;Get the next table entry
  OR   AL,AL           ;End of the Table?
  JZ  .O70             ;If so, error
  CMP  AL,AH           ;If not, is it a match?
  JNE .O10             ;If not, keep looking
.O20:                  ;POS is in the table
  TEST B [EMSFlags],UseFAT32 ;Do we support FAT32?
  JNZ .O80             ;If so, we're done
  CMP  AH,POSFAT32     ;If not, is it FAT32?
  JE  .O70             ;If so, we don't support it
  CMP  AH,POSFAT32LBA  ;Is it FAT32?
  JNE .O80             ;If not, it's OK
.O70:                  ;Error
  STC                  ;Set return flag
  JMP .O90             ;Done

.O80:                  ;OK
  CLC                  ;Set return flag
.O90:                  ;Done
  POP  SI,AX           ;Restore used registers
  RET

;------------------------------------------------------------------------------
;FILL IN DRIVE INFO ENTRY WITH PARAMETERS FOR NEW DRIVE
;Inputs:  DS = EMS/DPMS Segment/Selector
;         ES = TSR Segment/Selector
;         BX = Disk Info Entry
;         DI = Device Info Entry
;         SI = Drive Info Entry
;         All New Device Drive Parameters filled in
;Outputs: Copies New Device Parameters to Drive Parameters Section
;         Also copies NoFormat and NonFAT flags from Disk to Drive
;------------------------------------------------------------------------------
CopyNewDriveParams:
  PUSH EAX                       ;Save used registers
  MOV  EAX,[NewPartSector]       ;Copy
  MOV  D [SI+DrvPartSector+4],0  ;  Partition
  MOV  [SI+DrvPartSector],EAX    ;  Sector
  CMP  EAX,-1                    ;Is it a SuperFloppy?
  JNE .P10                       ;If not, continue
  MOV  [SI+DrvPartSector+4],EAX  ;If so, Store -1 in Total Partition Sector
.P10:                            ;Partition Sector Stored
  MOV  EAX,[NewBootSector]       ;Copy
  MOV  D [SI+DrvBootSector+4],0  ;  Boot
  MOV  [SI+DrvBootSector],EAX    ;  Sector
  MOV  EAX,[NewSectsInPart]      ;Copy
  MOV  D [SI+DrvSectsInPart+4],0 ;  Number
  MOV  [SI+DrvSectsInPart],EAX   ;  of Sectors
  MOV  AL,[NewPartOS]            ;Copy
  MOV  [SI+DrvPartOS],AL         ;  Partition Operating System
  MOV  AL,[NewPartIndex]         ;Copy
  MOV  [SI+DrvPartIndex],AL      ;  Partition Index
  CALL CopyFmtParams             ;Copy Format Parameters
.P20:                            ;Copy NoFormat flag from Disk to Drive
  TEST B [BX+DskFlags3],DskFlagNoFormat ;Unformatted Disk?
  JZ  .P30                       ;If not, continue
  OR   B [SI+DrvFlags],DrvFlagNoFormat ;If so, mark the Drive as Unformatted
.P30:                            ;Copy NonFAT flag from Disk to Drive
  TEST B [BX+DskFlags3],DskFlagNonFAT ;Non-FAT Disk?
  JZ  .P40                       ;If not, continue
  OR   B [SI+DrvFlags],DrvFlagNonFAT ;If so, mark the Drive as Non-FAT
.P40:                            ;Done
  POP  EAX                       ;Restore used registers
  RET

;------------------------------------------------------------------------------
;FILL IN DRIVE INFO ENTRY WITH PARAMETERS FOR NEW GPT DRIVE
;Inputs:  DS = EMS/DPMS Segment/Selector
;         ES = TSR Segment/Selector
;         BX = Disk Info Entry
;         DI = Device Info Entry
;         SI = Drive Info Entry
;         All New GPT Drive Parameters filled in
;Outputs: Copies New GPT Parameters to Drive Parameters Section
;------------------------------------------------------------------------------
CopyGPTDriveParams:
  PUSH EAX,EBX,ECX,EDX           ;Save used registers
  MOV  D [SI+DrvPartSector+4],0  ;Copy
  MOVZX EAX,W [GPTTblSector]     ;  Partition
  MOV  [SI+DrvPartSector],EAX    ;  Sector
  MOV  EDX,[LBAInBuffer+4]       ;Copy
  MOV  [SI+DrvBootSector+4],EDX  ;  Boot
  MOV  ECX,[LBAInBuffer]         ;  Sector
  MOV  [SI+DrvBootSector],ECX    ;  LBA
  MOV  EBX,[GPTLastPartSector+4] ;EBX:EAX =
  MOV  EAX,[GPTLastPartSector]   ;  Number
  SUB  EAX,ECX                   ;  of
  SBB  EBX,EDX                   ;  Sectors
  ADD  EAX,1                     ;  in the
  ADC  EBX,0                     ;  Partition
  MOV  [SI+DrvSectsInPart+4],EBX ;Store
  MOV  [SI+DrvSectsInPart],EAX   ;  it
  MOV  B [SI+DrvPartOS],POSGPT   ;Store Partition Operating System
  MOV  AL,[GPTCurrentSubCtr]     ;Copy
  MOV  [SI+DrvPartIndex],AL      ;  Partition Index
  CALL CopyFmtParams             ;Copy Format Parameters
  POP  EDX,ECX,EBX,EAX           ;Restore used registers
  RET

;------------------------------------------------------------------------------
;COPY THE PARAMETERS NEEDED FOR FORMATTING FROM THE DISK TO THE DRIVE
;Inputs:  DS = EMS/DPMS Segment/Selector
;         ES = TSR Segment/Selector
;         BX = Disk Info Entry
;         SI = Drive Info Entry
;Outputs: [SI].DrvFmtXX values
;------------------------------------------------------------------------------
CopyFmtParams:
  PUSH EAX,EDX                     ;Save used registers
  MOV  EDX,[BX+DskMaxLBA+4]        ;Copy
  MOV  EAX,[BX+DskMaxLBA]          ;  the
  ADD  EAX,1                       ;  Total
  ADC  EDX,0                       ;  Number
  MOV  [SI+DrvFmtTotalSects+4],EDX ;  of
  MOV  [SI+DrvFmtTotalSects],EAX   ;  Sectors
  MOV  AX,[BX+DskNumTracks]        ;Copy the
  MOV  [SI+DrvFmtCyls],AX          ;  Number of Cylinders
  MOV  AX,[BX+DskNumHeads]         ;Copy the
  MOV  [SI+DrvFmtHeads],AX         ;  Number of Heads
  MOV  AX,[BX+DskSectsPerTrack]    ;Copy the
  MOV  [SI+DrvFmtSects],AX         ;  Number of Sectors
  POP  EDX,EAX                     ;Restore used registers
  RET

;------------------------------------------------------------------------------
;FIND THE POINTER TO THE VOLUME NAME IN THE BOOT SECTOR
;Inputs:  DS = EMS/DPMS Segment/Selector
;         ES = TSR Segment/Selector
;         TSRDataSegment:SectorBuffer filled with Boot Sector Data
;           (contains BPB)
;Outputs: CF = Clear if OK (DOS 4+ Boot Sector)
;              TSRDataSegment:[SI] = Volume Name Pointer
;            = Set if Error (DOS 3- Boot Sector)
;              SI = Undefined
;------------------------------------------------------------------------------
GetVolNamePtr:
  PUSH AX,DX,DS        ;Save used registers
  MOV  DX,SI           ;Save original SI
  CALL GetTSRDataSegDS ;DS = TSR Data Segment
.P10:                  ;Test for FAT32
  MOV  SI,(SectorBuffer+11+BPBXVolumeName)
                       ;Assume FAT32
  CMP  B [SectorBuffer+BPBXSignature+11],29h
                       ;Possibly FAT32?
  JNE .P20             ;If not, try next possibility
  CALL TestVolumeName  ;Test the Volume Name for validity
  JNC .P80             ;If OK, we're done
.P20:                  ;Test for FAT12/FAT16 DOS 4+
  MOV  SI,(SectorBuffer+11+BPB4VolumeName)
                       ;Assume FAT12/FAT16 DOS 4+
  MOV  AL,[SectorBuffer+BPB4Signature+11]
                       ;AL = Signature Byte
  CMP  AL,28h          ;Possibly DOS 4+?
  JB  .P70             ;If not, it's DOS 3-
  CMP  AL,29h          ;Possibly DOS 4+?
  JA  .P70             ;If not, it's DOS 3-
  CALL TestVolumeName  ;Test the Volume Name for validity
  JNC .P80             ;If OK, we're done
.P70:                  ;Error
  MOV  SI,DX           ;Restore original SI
  STC                  ;Set return flag
  JMP .P90             ;Done

.P80:                  ;OK
  CLC                  ;Set return flag
.P90:                  ;Done
  POP  DS,DX,AX        ;Restore used registers
  RET

;------------------------------------------------------------------------------
;TEST A POTENTIAL VOLUME NAME TO SEE IF IT'S VALID
;Inputs:  DS:SI = Potential Volume Name
;         CLD already issued
;Outputs: CF = Clear if OK (appears to be a valid Volume Name)
;            = Set if Error
;------------------------------------------------------------------------------
TestVolumeName:
  PUSH AX,CX  ;Save used registers
  PUSH SI     ;Save used registers
  MOV  CX,11  ;Test 11 bytes
.V10:         ;Loop to here for each character
  LODSB       ;Get the next character
  CMP  AL,' ' ;Valid ASCII character?
  JB  .V70    ;If not, invalid string
  CMP  AL,127 ;Valid ASCII character? 8-bit characters are okay, 127 is not.
  JE  .V70    ;If not, invalid string
  LOOP .V10   ;Keep going until we're done
  JMP .V80    ;All characters appear valid

.V70:         ;Error
  STC         ;Set return flag
  JMP .V90    ;Done

.V80:         ;OK
  CLC         ;Set return flag
.V90:         ;Done
  POP  SI     ;Restore used registers
  POP  CX,AX  ;Restore used registers
  RET

;------------------------------------------------------------------------------
;ISSUE A REQUEST TO READ A SECTOR FROM A DISK AS A BACKGROUND PROCESS
;Inputs:  DS = EMS/DPMS Segment/Selector
;         ES = TSR Segment/Selector
;         BX = Disk Info Entry
;         DI = Device Info Entry
;         [BX].DskFlags2.DskFlagLBA64 set if we need to issue LBA64 Request
;         EDX:EAX = Sector Number to Read
;Outputs: Sends SCSI Request
;         CF = Clear if it worked OK
;              Appropriate Buffer cleared, then filled with Data
;            = Set if Error
;Changes:
;------------------------------------------------------------------------------
SCSIReadSectorEAX0BG:
  PUSH EDX              ;Save used registers
  XOR  EDX,EDX          ;EDX = 0
  CALL SCSIReadSectorBG ;Read the Sector
  POP  EDX              ;Restore used registers
  RET

SCSIReadSectorEAXBG:
	; actual USB specific code omitted, sector number EAX
  RET

SCSIReadSectorBG:
	; either call SCSIReadSectorEAXBG (ignoring EDX) or, only
	; if 64-bit supported, use 64-bit sector number EDX:EAX
  RET


;------------------------------------------------------------------------------
; To manually trigger processing of a drive, call ProcessDrive,
; which is not from the original code. The USB driver would read
; the first sector as part of the detection process and set the
; callback to GotPartSector... (when queueing the read) instead.
; Inputs:  None
; Outputs: None
; Changes: All registers
;------------------------------------------------------------------------------
ProcessDrive:
  CALL SetupRETFRegs2        ;DS:BX = Disk, DS:DI = Device,
                             ;ES:SI = Int14Request, AX = Error
                             ;GS = TSR Data Segment/Selector
  ; MOV  B [NewDvcStage],1   ;Update Stage (status, for debugging only)
  JNZ .E70                   ;If error, quit
  ; MOV  B [NewDvcStage],2   ;Update Stage (status, for debugging only)
  ; MOV  W [NextSCSIReqOfstRM],GotPartSectorXX ;Go here
  ; MOV  W [NextSCSIReqOfstPM],GotPartSectorPM ;  when done
  MOV  W [NextSCSIReqOfst],GotPartSectorCode ; go here when SCSI I/O complete
  XOR  EAX,EAX                ;Read Sector 0
  CALL SCSIReadSectorEAX0BG   ;Issue the Read Sector Request for sector EAX
  JC  .E70                    ;If error, quit
  CALL SetupSCSIDone          ;Issue the Request
  JNC .E90                    ;If no error, we are done

.E70:                        ;Error
  CALL FinishUpBadDrive      ;Mark it as Bad
  JMP .E90                   ;Done
.E90:
  RETF
